from __future__ import annotations

import warnings
from collections.abc import Iterable
from typing import Any


__all__ = ["lazy_import"]


def import_name(name: str, source: str, namespace: dict[str, Any]) -> Any:
    """
    Import ``name`` from ``source`` in ``namespace``.

    There are two use cases:

    - ``name`` is an object defined in ``source``;
    - ``name`` is a submodule of ``source``.

    Neither :func:`__import__` nor :func:`~importlib.import_module` does
    exactly this. :func:`__import__` is closer to the intended behavior.

    """
    level = 0
    while source[level] == ".":
        level += 1
        assert level < len(source), "importing from parent isn't supported"
    module = __import__(source[level:], namespace, None, [name], level)
    return getattr(module, name)


def lazy_import(
    namespace: dict[str, Any],
    aliases: dict[str, str] | None = None,
    deprecated_aliases: dict[str, str] | None = None,
) -> None:
    """
    Provide lazy, module-level imports.

    Typical use::

        __getattr__, __dir__ = lazy_import(
            globals(),
            aliases={
                "<name>": "<source module>",
                ...
            },
            deprecated_aliases={
                ...,
            }
        )

    This function defines ``__getattr__`` and ``__dir__`` per :pep:`562`.

    """
    if aliases is None:
        aliases = {}
    if deprecated_aliases is None:
        deprecated_aliases = {}

    namespace_set = set(namespace)
    aliases_set = set(aliases)
    deprecated_aliases_set = set(deprecated_aliases)

    assert not namespace_set & aliases_set, "namespace conflict"
    assert not namespace_set & deprecated_aliases_set, "namespace conflict"
    assert not aliases_set & deprecated_aliases_set, "namespace conflict"

    package = namespace["__name__"]

    def __getattr__(name: str) -> Any:
        assert aliases is not None  # mypy cannot figure this out
        try:
            source = aliases[name]
        except KeyError:
            pass
        else:
            return import_name(name, source, namespace)

        assert deprecated_aliases is not None  # mypy cannot figure this out
        try:
            source = deprecated_aliases[name]
        except KeyError:
            pass
        else:
            warnings.warn(
                f"{package}.{name} is deprecated",
                DeprecationWarning,
                stacklevel=2,
            )
            return import_name(name, source, namespace)

        raise AttributeError(f"module {package!r} has no attribute {name!r}")

    namespace["__getattr__"] = __getattr__

    def __dir__() -> Iterable[str]:
        return sorted(namespace_set | aliases_set | deprecated_aliases_set)

    namespace["__dir__"] = __dir__



#!/usr/bin/env python

"""Client using the threading API."""

from websockets.sync.client import connect


def hello():
    with connect("ws://localhost:8765") as websocket:
        websocket.send("Hello world!")
        message = websocket.recv()
        print(message)


if __name__ == "__main__":
    hello()



import logging
import os
import platform


if platform.python_implementation() != "PyPy":  # pragma: no branch
    import tracemalloc


format = "%(asctime)s %(levelname)s %(name)s %(message)s"

if bool(os.environ.get("WEBSOCKETS_DEBUG")):  # pragma: no cover
    # Display every frame sent or received in debug mode.
    level = logging.DEBUG
else:
    # Hide stack traces of exceptions.
    level = logging.CRITICAL

logging.basicConfig(format=format, level=level)

if bool(os.environ.get("WEBSOCKETS_TRACE")):  # pragma: no cover
    # Trace allocations to debug resource warnings.
    tracemalloc.start()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import datetime
import importlib
import inspect
import os
import subprocess
import sys

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.join(os.path.abspath(".."), "src"))


# -- Project information -----------------------------------------------------

project = "websockets"
copyright = f"2013-{datetime.date.today().year}, Aymeric Augustin and contributors"
author = "Aymeric Augustin"

from websockets.version import tag as version, version as release


# -- General configuration ---------------------------------------------------

nitpicky = True

nitpick_ignore = [
    # topics/design.rst discusses undocumented APIs
    ("py:meth", "client.WebSocketClientProtocol.handshake"),
    ("py:meth", "server.WebSocketServerProtocol.handshake"),
    ("py:attr", "protocol.WebSocketCommonProtocol.is_client"),
    ("py:attr", "protocol.WebSocketCommonProtocol.messages"),
    ("py:meth", "protocol.WebSocketCommonProtocol.close_connection"),
    ("py:attr", "protocol.WebSocketCommonProtocol.close_connection_task"),
    ("py:meth", "protocol.WebSocketCommonProtocol.keepalive_ping"),
    ("py:attr", "protocol.WebSocketCommonProtocol.keepalive_ping_task"),
    ("py:meth", "protocol.WebSocketCommonProtocol.transfer_data"),
    ("py:attr", "protocol.WebSocketCommonProtocol.transfer_data_task"),
    ("py:meth", "protocol.WebSocketCommonProtocol.connection_open"),
    ("py:meth", "protocol.WebSocketCommonProtocol.ensure_open"),
    ("py:meth", "protocol.WebSocketCommonProtocol.fail_connection"),
    ("py:meth", "protocol.WebSocketCommonProtocol.connection_lost"),
    ("py:meth", "protocol.WebSocketCommonProtocol.read_message"),
    ("py:meth", "protocol.WebSocketCommonProtocol.write_frame"),
]

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "sphinx_copybutton",
    "sphinx_inline_tabs",
    "sphinxcontrib.spelling",
    "sphinxcontrib_trio",
    "sphinxext.opengraph",
]
# It is currently inconvenient to install PyEnchant on Apple Silicon.
try:
    import sphinxcontrib.spelling
except ImportError:
    extensions.remove("sphinxcontrib.spelling")

autodoc_typehints = "description"

autodoc_typehints_description_target = "documented"

# Workaround for https://github.com/sphinx-doc/sphinx/issues/9560
from sphinx.domains.python import PythonDomain

assert PythonDomain.object_types["data"].roles == ("data", "obj")
PythonDomain.object_types["data"].roles = ("data", "class", "obj")

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "sesame": ("https://django-sesame.readthedocs.io/en/stable/", None),
    "werkzeug": ("https://werkzeug.palletsprojects.com/en/stable/", None),
}

spelling_show_suggestions = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# Configure viewcode extension.
from websockets.version import commit

code_url = f"https://github.com/python-websockets/websockets/blob/{commit}"

def linkcode_resolve(domain, info):
    # Non-linkable objects from the starter kit in the tutorial.
    if domain == "js" or info["module"] == "connect4":
        return

    assert domain == "py", "expected only Python objects"

    mod = importlib.import_module(info["module"])
    if "." in info["fullname"]:
        objname, attrname = info["fullname"].split(".")
        obj = getattr(mod, objname)
        try:
            # object is a method of a class
            obj = getattr(obj, attrname)
        except AttributeError:
            # object is an attribute of a class
            return None
    else:
        obj = getattr(mod, info["fullname"])

    try:
        file = inspect.getsourcefile(obj)
        lines = inspect.getsourcelines(obj)
    except TypeError:
        # e.g. object is a typing.Union
        return None
    file = os.path.relpath(file, os.path.abspath(".."))
    if not file.startswith("src/websockets"):
        # e.g. object is a typing.NewType
        return None
    start, end = lines[1], lines[1] + len(lines[0]) - 1

    return f"{code_url}/{file}#L{start}-L{end}"

# Configure opengraph extension

# Social cards don't support the SVG logo. Also, the text preview looks bad.
ogp_social_cards = {"enable": False}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "furo"

html_theme_options = {
    "light_css_variables": {
        "color-brand-primary": "#306998",  # blue from logo
        "color-brand-content": "#0b487a",  # blue more saturated and less dark
    },
    "dark_css_variables": {
        "color-brand-primary": "#ffd43bcc",  # yellow from logo, more muted than content
        "color-brand-content": "#ffd43bd9",  # yellow from logo, transparent like text
    },
    "sidebar_hide_name": True,
}

html_logo = "_static/websockets.svg"

html_favicon = "_static/favicon.ico"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_copy_source = False

html_show_sphinx = False



from __future__ import annotations

import warnings

from .datastructures import Headers, MultipleValuesError  # noqa: F401


with warnings.catch_warnings():
    # Suppress redundant DeprecationWarning raised by websockets.legacy.
    warnings.filterwarnings("ignore", category=DeprecationWarning)
    from .legacy.http import read_request, read_response  # noqa: F401


warnings.warn(  # deprecated in 9.0 - 2021-09-01
    "Headers and MultipleValuesError were moved "
    "from websockets.http to websockets.datastructures"
    "and read_request and read_response were moved "
    "from websockets.http to websockets.legacy.http",
    DeprecationWarning,
)



import http
import socket
import sys
import unittest
from unittest.mock import patch

from websockets.asyncio.client import connect, unix_connect
from websockets.asyncio.router import *
from websockets.exceptions import InvalidStatus

from ..utils import CLIENT_CONTEXT, SERVER_CONTEXT, alist, temp_unix_socket_path
from .server import EvalShellMixin, get_uri, handler


try:
    from werkzeug.routing import Map, Rule
except ImportError:
    pass


async def echo(websocket, count):
    message = await websocket.recv()
    for _ in range(count):
        await websocket.send(message)


@unittest.skipUnless("werkzeug" in sys.modules, "werkzeug not installed")
class RouterTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    # This is a small realistic example of werkzeug's basic URL routing
    # features: path matching, parameter extraction, and default values.

    async def test_router_matches_paths_and_extracts_parameters(self):
        """Router matches paths and extracts parameters."""
        url_map = Map(
            [
                Rule("/echo", defaults={"count": 1}, endpoint=echo),
                Rule("/echo/<int:count>", endpoint=echo),
            ]
        )
        async with route(url_map, "localhost", 0) as server:
            async with connect(get_uri(server) + "/echo") as client:
                await client.send("hello")
                messages = await alist(client)
            self.assertEqual(messages, ["hello"])

            async with connect(get_uri(server) + "/echo/3") as client:
                await client.send("hello")
                messages = await alist(client)
            self.assertEqual(messages, ["hello", "hello", "hello"])

    @property  # avoids an import-time dependency on werkzeug
    def url_map(self):
        return Map(
            [
                Rule("/", endpoint=handler),
                Rule("/r", redirect_to="/"),
            ]
        )

    async def test_route_with_query_string(self):
        """Router ignores query strings when matching paths."""
        async with route(self.url_map, "localhost", 0) as server:
            async with connect(get_uri(server) + "/?a=b") as client:
                await self.assertEval(client, "ws.request.path", "/?a=b")

    async def test_redirect(self):
        """Router redirects connections according to redirect_to."""
        async with route(self.url_map, "localhost", 0) as server:
            async with connect(get_uri(server) + "/r") as client:
                await self.assertEval(client, "ws.request.path", "/")

    async def test_secure_redirect(self):
        """Router redirects connections to a wss:// URI when TLS is enabled."""
        async with route(self.url_map, "localhost", 0, ssl=SERVER_CONTEXT) as server:
            async with connect(get_uri(server) + "/r", ssl=CLIENT_CONTEXT) as client:
                await self.assertEval(client, "ws.request.path", "/")

    @patch("websockets.asyncio.client.connect.process_redirect", lambda _, exc: exc)
    async def test_force_secure_redirect(self):
        """Router redirects ws:// connections to a wss:// URI when ssl=True."""
        async with route(self.url_map, "localhost", 0, ssl=True) as server:
            redirect_uri = get_uri(server, secure=True)
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server) + "/r"):
                    self.fail("did not raise")
        self.assertEqual(
            raised.exception.response.headers["Location"],
            redirect_uri + "/",
        )

    @patch("websockets.asyncio.client.connect.process_redirect", lambda _, exc: exc)
    async def test_force_redirect_server_name(self):
        """Router redirects connections to the host declared in server_name."""
        async with route(self.url_map, "localhost", 0, server_name="other") as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server) + "/r"):
                    self.fail("did not raise")
        self.assertEqual(
            raised.exception.response.headers["Location"],
            "ws://other/",
        )

    async def test_not_found(self):
        """Router rejects requests to unknown paths with an HTTP 404 error."""
        async with route(self.url_map, "localhost", 0) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server) + "/n"):
                    self.fail("did not raise")
        self.assertEqual(
            str(raised.exception),
            "server rejected WebSocket connection: HTTP 404",
        )

    async def test_process_request_function_returning_none(self):
        """Router supports a process_request function returning None."""

        def process_request(ws, request):
            ws.process_request_ran = True

        async with route(
            self.url_map, "localhost", 0, process_request=process_request
        ) as server:
            async with connect(get_uri(server) + "/") as client:
                await self.assertEval(client, "ws.process_request_ran", "True")

    async def test_process_request_coroutine_returning_none(self):
        """Router supports a process_request coroutine returning None."""

        async def process_request(ws, request):
            ws.process_request_ran = True

        async with route(
            self.url_map, "localhost", 0, process_request=process_request
        ) as server:
            async with connect(get_uri(server) + "/") as client:
                await self.assertEval(client, "ws.process_request_ran", "True")

    async def test_process_request_function_returning_response(self):
        """Router supports a process_request function returning a response."""

        def process_request(ws, request):
            return ws.respond(http.HTTPStatus.FORBIDDEN, "Forbidden")

        async with route(
            self.url_map, "localhost", 0, process_request=process_request
        ) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server) + "/"):
                    self.fail("did not raise")
        self.assertEqual(
            str(raised.exception),
            "server rejected WebSocket connection: HTTP 403",
        )

    async def test_process_request_coroutine_returning_response(self):
        """Router supports a process_request coroutine returning a response."""

        async def process_request(ws, request):
            return ws.respond(http.HTTPStatus.FORBIDDEN, "Forbidden")

        async with route(
            self.url_map, "localhost", 0, process_request=process_request
        ) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server) + "/"):
                    self.fail("did not raise")
        self.assertEqual(
            str(raised.exception),
            "server rejected WebSocket connection: HTTP 403",
        )

    async def test_custom_router_factory(self):
        """Router supports a custom router factory."""

        class MyRouter(Router):
            async def handler(self, connection):
                connection.my_router_ran = True
                return await super().handler(connection)

        async with route(
            self.url_map, "localhost", 0, create_router=MyRouter
        ) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.my_router_ran", "True")


@unittest.skipUnless(hasattr(socket, "AF_UNIX"), "this test requires Unix sockets")
@unittest.skipUnless("werkzeug" in sys.modules, "werkzeug not installed")
class UnixRouterTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    async def test_router_supports_unix_sockets(self):
        """Router supports Unix sockets."""
        url_map = Map([Rule("/echo/<int:count>", endpoint=echo)])
        with temp_unix_socket_path() as path:
            async with unix_route(url_map, path):
                async with unix_connect(path, "ws://localhost/echo/3") as client:
                    await client.send("hello")
                    messages = await alist(client)
                self.assertEqual(messages, ["hello", "hello", "hello"])



import asyncio
import contextlib
import itertools
import logging
import socket
import sys
import unittest
import uuid
from unittest.mock import Mock, patch

from websockets.asyncio.compatibility import TimeoutError, aiter, anext, asyncio_timeout
from websockets.asyncio.connection import *
from websockets.asyncio.connection import broadcast
from websockets.exceptions import (
    ConcurrencyError,
    ConnectionClosedError,
    ConnectionClosedOK,
)
from websockets.frames import CloseCode, Frame, Opcode
from websockets.protocol import CLIENT, SERVER, Protocol, State

from ..protocol import RecordingProtocol
from ..utils import MS, alist
from .connection import InterceptingConnection


# Connection implements symmetrical behavior between clients and servers.
# All tests run on the client side and the server side to validate this.


class ClientConnectionTests(unittest.IsolatedAsyncioTestCase):
    LOCAL = CLIENT
    REMOTE = SERVER

    async def asyncSetUp(self):
        self.loop = asyncio.get_running_loop()
        socket_, remote_socket = socket.socketpair()
        protocol = Protocol(self.LOCAL)
        remote_protocol = RecordingProtocol(self.REMOTE)
        self.transport, self.connection = await self.loop.create_connection(
            lambda: Connection(protocol, close_timeout=2 * MS),
            sock=socket_,
        )
        _remote_transport, self.remote_connection = await self.loop.create_connection(
            lambda: InterceptingConnection(remote_protocol),
            sock=remote_socket,
        )

    async def asyncTearDown(self):
        await self.remote_connection.close()
        await self.connection.close()

    # Test helpers built upon RecordingProtocol and InterceptingConnection.

    async def wait_for_remote_side(self):
        """Wait for the remote side to process messages."""
        # Two runs of the event loop are required for answering pings.
        await asyncio.sleep(0)
        await asyncio.sleep(0)

    async def assertFrameSent(self, frame):
        """Check that a single frame was sent."""
        await self.wait_for_remote_side()
        self.assertEqual(self.remote_connection.protocol.get_frames_rcvd(), [frame])

    async def assertFramesSent(self, frames):
        """Check that several frames were sent."""
        await self.wait_for_remote_side()
        self.assertEqual(self.remote_connection.protocol.get_frames_rcvd(), frames)

    async def assertNoFrameSent(self):
        """Check that no frame was sent."""
        await self.wait_for_remote_side()
        self.assertEqual(self.remote_connection.protocol.get_frames_rcvd(), [])

    @contextlib.asynccontextmanager
    async def delay_frames_rcvd(self, delay):
        """Delay frames before they're received by the connection."""
        with self.remote_connection.delay_frames_sent(delay):
            yield
            await self.wait_for_remote_side()

    @contextlib.asynccontextmanager
    async def delay_eof_rcvd(self, delay):
        """Delay EOF before it's received by the connection."""
        with self.remote_connection.delay_eof_sent(delay):
            yield
            await self.wait_for_remote_side()

    @contextlib.asynccontextmanager
    async def drop_frames_rcvd(self):
        """Drop frames before they're received by the connection."""
        with self.remote_connection.drop_frames_sent():
            yield
            await self.wait_for_remote_side()

    @contextlib.asynccontextmanager
    async def drop_eof_rcvd(self):
        """Drop EOF before it's received by the connection."""
        with self.remote_connection.drop_eof_sent():
            yield
            await self.wait_for_remote_side()

    # Test __aenter__ and __aexit__.

    async def test_aenter(self):
        """__aenter__ returns the connection itself."""
        async with self.connection as connection:
            self.assertIs(connection, self.connection)

    async def test_aexit(self):
        """__aexit__ closes the connection with code 1000."""
        async with self.connection:
            await self.assertNoFrameSent()
        await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

    async def test_aexit_with_exception(self):
        """__aexit__ with an exception closes the connection with code 1011."""
        with self.assertRaises(RuntimeError):
            async with self.connection:
                raise RuntimeError
        await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xf3"))

    # Test __aiter__.

    async def test_aiter_text(self):
        """__aiter__ yields text messages."""
        iterator = aiter(self.connection)
        async with contextlib.aclosing(iterator):
            await self.remote_connection.send("😀")
            self.assertEqual(await anext(iterator), "😀")
            await self.remote_connection.send("😀")
            self.assertEqual(await anext(iterator), "😀")

    async def test_aiter_binary(self):
        """__aiter__ yields binary messages."""
        iterator = aiter(self.connection)
        async with contextlib.aclosing(iterator):
            await self.remote_connection.send(b"\x01\x02\xfe\xff")
            self.assertEqual(await anext(iterator), b"\x01\x02\xfe\xff")
            await self.remote_connection.send(b"\x01\x02\xfe\xff")
            self.assertEqual(await anext(iterator), b"\x01\x02\xfe\xff")

    async def test_aiter_mixed(self):
        """__aiter__ yields a mix of text and binary messages."""
        iterator = aiter(self.connection)
        async with contextlib.aclosing(iterator):
            await self.remote_connection.send("😀")
            self.assertEqual(await anext(iterator), "😀")
            await self.remote_connection.send(b"\x01\x02\xfe\xff")
            self.assertEqual(await anext(iterator), b"\x01\x02\xfe\xff")

    async def test_aiter_connection_closed_ok(self):
        """__aiter__ terminates after a normal closure."""
        iterator = aiter(self.connection)
        async with contextlib.aclosing(iterator):
            await self.remote_connection.close()
            with self.assertRaises(StopAsyncIteration):
                await anext(iterator)

    async def test_aiter_connection_closed_error(self):
        """__aiter__ raises ConnectionClosedError after an error."""
        iterator = aiter(self.connection)
        async with contextlib.aclosing(iterator):
            await self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
            with self.assertRaises(ConnectionClosedError):
                await anext(iterator)

    # Test recv.

    async def test_recv_text(self):
        """recv receives a text message."""
        await self.remote_connection.send("😀")
        self.assertEqual(await self.connection.recv(), "😀")

    async def test_recv_binary(self):
        """recv receives a binary message."""
        await self.remote_connection.send(b"\x01\x02\xfe\xff")
        self.assertEqual(await self.connection.recv(), b"\x01\x02\xfe\xff")

    async def test_recv_text_as_bytes(self):
        """recv receives a text message as bytes."""
        await self.remote_connection.send("😀")
        self.assertEqual(await self.connection.recv(decode=False), "😀".encode())

    async def test_recv_binary_as_text(self):
        """recv receives a binary message as a str."""
        await self.remote_connection.send("😀".encode())
        self.assertEqual(await self.connection.recv(decode=True), "😀")

    async def test_recv_fragmented_text(self):
        """recv receives a fragmented text message."""
        await self.remote_connection.send(["😀", "😀"])
        self.assertEqual(await self.connection.recv(), "😀😀")

    async def test_recv_fragmented_binary(self):
        """recv receives a fragmented binary message."""
        await self.remote_connection.send([b"\x01\x02", b"\xfe\xff"])
        self.assertEqual(await self.connection.recv(), b"\x01\x02\xfe\xff")

    async def test_recv_connection_closed_ok(self):
        """recv raises ConnectionClosedOK after a normal closure."""
        await self.remote_connection.close()
        with self.assertRaises(ConnectionClosedOK):
            await self.connection.recv()

    async def test_recv_connection_closed_error(self):
        """recv raises ConnectionClosedError after an error."""
        await self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
        with self.assertRaises(ConnectionClosedError):
            await self.connection.recv()

    async def test_recv_non_utf8_text(self):
        """recv receives a non-UTF-8 text message."""
        await self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
        with self.assertRaises(ConnectionClosedError) as raised:
            await self.connection.recv()
        self.assertEqual(raised.exception.sent.code, CloseCode.INVALID_DATA)

    async def test_recv_during_recv(self):
        """recv raises ConcurrencyError when called concurrently."""
        recv_task = asyncio.create_task(self.connection.recv())
        await asyncio.sleep(0)  # let the event loop start recv_task
        try:
            with self.assertRaises(ConcurrencyError) as raised:
                await self.connection.recv()
        finally:
            recv_task.cancel()
        self.assertEqual(
            str(raised.exception),
            "cannot call recv while another coroutine "
            "is already running recv or recv_streaming",
        )

    async def test_recv_during_recv_streaming(self):
        """recv raises ConcurrencyError when called concurrently with recv_streaming."""
        recv_streaming_task = asyncio.create_task(
            alist(self.connection.recv_streaming())
        )
        await asyncio.sleep(0)  # let the event loop start recv_streaming_task
        try:
            with self.assertRaises(ConcurrencyError) as raised:
                await self.connection.recv()
        finally:
            recv_streaming_task.cancel()
        self.assertEqual(
            str(raised.exception),
            "cannot call recv while another coroutine "
            "is already running recv or recv_streaming",
        )

    async def test_recv_cancellation_before_receiving(self):
        """recv can be canceled before receiving a message."""
        recv_task = asyncio.create_task(self.connection.recv())
        await asyncio.sleep(0)  # let the event loop start recv_task
        recv_task.cancel()
        await asyncio.sleep(0)  # let the event loop cancel recv_task

        # Running recv again receives the next message.
        await self.remote_connection.send("😀")
        self.assertEqual(await self.connection.recv(), "😀")

    async def test_recv_cancellation_while_receiving(self):
        """recv can be canceled while receiving a fragmented message."""
        gate = asyncio.Event()

        async def fragments():
            yield "⏳"
            await gate.wait()
            yield "⌛️"

        asyncio.create_task(self.remote_connection.send(fragments()))
        await asyncio.sleep(0)

        recv_task = asyncio.create_task(self.connection.recv())
        await asyncio.sleep(0)  # let the event loop start recv_task
        recv_task.cancel()
        await asyncio.sleep(0)  # let the event loop cancel recv_task

        gate.set()

        # Running recv again receives the complete message.
        self.assertEqual(await self.connection.recv(), "⏳⌛️")

    # Test recv_streaming.

    async def test_recv_streaming_text(self):
        """recv_streaming receives a text message."""
        await self.remote_connection.send("😀")
        self.assertEqual(
            await alist(self.connection.recv_streaming()),
            ["😀"],
        )

    async def test_recv_streaming_binary(self):
        """recv_streaming receives a binary message."""
        await self.remote_connection.send(b"\x01\x02\xfe\xff")
        self.assertEqual(
            await alist(self.connection.recv_streaming()),
            [b"\x01\x02\xfe\xff"],
        )

    async def test_recv_streaming_text_as_bytes(self):
        """recv_streaming receives a text message as bytes."""
        await self.remote_connection.send("😀")
        self.assertEqual(
            await alist(self.connection.recv_streaming(decode=False)),
            ["😀".encode()],
        )

    async def test_recv_streaming_binary_as_str(self):
        """recv_streaming receives a binary message as a str."""
        await self.remote_connection.send("😀".encode())
        self.assertEqual(
            await alist(self.connection.recv_streaming(decode=True)),
            ["😀"],
        )

    async def test_recv_streaming_fragmented_text(self):
        """recv_streaming receives a fragmented text message."""
        await self.remote_connection.send(["😀", "😀"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    async def test_recv_streaming_fragmented_binary(self):
        """recv_streaming receives a fragmented binary message."""
        await self.remote_connection.send([b"\x01\x02", b"\xfe\xff"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.connection.recv_streaming()),
            [b"\x01\x02", b"\xfe\xff", b""],
        )

    async def test_recv_streaming_connection_closed_ok(self):
        """recv_streaming raises ConnectionClosedOK after a normal closure."""
        await self.remote_connection.close()
        with self.assertRaises(ConnectionClosedOK):
            async for _ in self.connection.recv_streaming():
                self.fail("did not raise")

    async def test_recv_streaming_connection_closed_error(self):
        """recv_streaming raises ConnectionClosedError after an error."""
        await self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
        with self.assertRaises(ConnectionClosedError):
            async for _ in self.connection.recv_streaming():
                self.fail("did not raise")

    async def test_recv_streaming_non_utf8_text(self):
        """recv_streaming receives a non-UTF-8 text message."""
        await self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
        with self.assertRaises(ConnectionClosedError) as raised:
            await alist(self.connection.recv_streaming())
        self.assertEqual(raised.exception.sent.code, CloseCode.INVALID_DATA)

    async def test_recv_streaming_during_recv(self):
        """recv_streaming raises ConcurrencyError when called concurrently with recv."""
        recv_task = asyncio.create_task(self.connection.recv())
        await asyncio.sleep(0)  # let the event loop start recv_task
        try:
            with self.assertRaises(ConcurrencyError) as raised:
                async for _ in self.connection.recv_streaming():
                    self.fail("did not raise")
        finally:
            recv_task.cancel()
        self.assertEqual(
            str(raised.exception),
            "cannot call recv_streaming while another coroutine "
            "is already running recv or recv_streaming",
        )

    async def test_recv_streaming_during_recv_streaming(self):
        """recv_streaming raises ConcurrencyError when called concurrently."""
        recv_streaming_task = asyncio.create_task(
            alist(self.connection.recv_streaming())
        )
        await asyncio.sleep(0)  # let the event loop start recv_streaming_task
        try:
            with self.assertRaises(ConcurrencyError) as raised:
                async for _ in self.connection.recv_streaming():
                    self.fail("did not raise")
        finally:
            recv_streaming_task.cancel()
        self.assertEqual(
            str(raised.exception),
            "cannot call recv_streaming while another coroutine "
            "is already running recv or recv_streaming",
        )

    async def test_recv_streaming_cancellation_before_receiving(self):
        """recv_streaming can be canceled before receiving a message."""
        recv_streaming_task = asyncio.create_task(
            alist(self.connection.recv_streaming())
        )
        await asyncio.sleep(0)  # let the event loop start recv_streaming_task
        recv_streaming_task.cancel()
        await asyncio.sleep(0)  # let the event loop cancel recv_streaming_task

        # Running recv_streaming again receives the next message.
        await self.remote_connection.send(["😀", "😀"])
        self.assertEqual(
            await alist(self.connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    async def test_recv_streaming_cancellation_while_receiving(self):
        """recv_streaming cannot be canceled while receiving a fragmented message."""
        gate = asyncio.Event()

        async def fragments():
            yield "⏳"
            await gate.wait()
            yield "⌛️"

        asyncio.create_task(self.remote_connection.send(fragments()))
        await asyncio.sleep(0)

        recv_streaming_task = asyncio.create_task(
            alist(self.connection.recv_streaming())
        )
        await asyncio.sleep(0)  # let the event loop start recv_streaming_task
        await asyncio.sleep(0)  # experimentally, two runs of the event loop
        await asyncio.sleep(0)  # are needed to receive the first fragment
        recv_streaming_task.cancel()
        await asyncio.sleep(0)  # let the event loop cancel recv_streaming_task

        gate.set()
        await asyncio.sleep(0)

        # Running recv_streaming again fails.
        with self.assertRaises(ConcurrencyError):
            async for _ in self.connection.recv_streaming():
                self.fail("did not raise")

    # Test send.

    async def test_send_text(self):
        """send sends a text message."""
        await self.connection.send("😀")
        self.assertEqual(await self.remote_connection.recv(), "😀")

    async def test_send_binary(self):
        """send sends a binary message."""
        await self.connection.send(b"\x01\x02\xfe\xff")
        self.assertEqual(await self.remote_connection.recv(), b"\x01\x02\xfe\xff")

    async def test_send_binary_from_str(self):
        """send sends a binary message from a str."""
        await self.connection.send("😀", text=False)
        self.assertEqual(await self.remote_connection.recv(), "😀".encode())

    async def test_send_text_from_bytes(self):
        """send sends a text message from bytes."""
        await self.connection.send("😀".encode(), text=True)
        self.assertEqual(await self.remote_connection.recv(), "😀")

    async def test_send_fragmented_text(self):
        """send sends a fragmented text message."""
        await self.connection.send(["😀", "😀"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    async def test_send_fragmented_binary(self):
        """send sends a fragmented binary message."""
        await self.connection.send([b"\x01\x02", b"\xfe\xff"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            [b"\x01\x02", b"\xfe\xff", b""],
        )

    async def test_send_fragmented_binary_from_str(self):
        """send sends a fragmented binary message from a str."""
        await self.connection.send(["😀", "😀"], text=False)
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            ["😀".encode(), "😀".encode(), b""],
        )

    async def test_send_fragmented_text_from_bytes(self):
        """send sends a fragmented text message from bytes."""
        await self.connection.send(["😀".encode(), "😀".encode()], text=True)
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    async def test_send_async_fragmented_text(self):
        """send sends a fragmented text message asynchronously."""

        async def fragments():
            yield "😀"
            yield "😀"

        await self.connection.send(fragments())
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    async def test_send_async_fragmented_binary(self):
        """send sends a fragmented binary message asynchronously."""

        async def fragments():
            yield b"\x01\x02"
            yield b"\xfe\xff"

        await self.connection.send(fragments())
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            [b"\x01\x02", b"\xfe\xff", b""],
        )

    async def test_send_async_fragmented_binary_from_str(self):
        """send sends a fragmented binary message from a str asynchronously."""

        async def fragments():
            yield "😀"
            yield "😀"

        await self.connection.send(fragments(), text=False)
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            ["😀".encode(), "😀".encode(), b""],
        )

    async def test_send_async_fragmented_text_from_bytes(self):
        """send sends a fragmented text message from bytes asynchronously."""

        async def fragments():
            yield "😀".encode()
            yield "😀".encode()

        await self.connection.send(fragments(), text=True)
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            await alist(self.remote_connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    async def test_send_connection_closed_ok(self):
        """send raises ConnectionClosedOK after a normal closure."""
        await self.remote_connection.close()
        with self.assertRaises(ConnectionClosedOK):
            await self.connection.send("😀")

    async def test_send_connection_closed_error(self):
        """send raises ConnectionClosedError after an error."""
        await self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
        with self.assertRaises(ConnectionClosedError):
            await self.connection.send("😀")

    async def test_send_during_send(self):
        """send waits for a previous call to send to complete."""
        # This test fails if the guard with send_in_progress is removed
        # from send() in the case when message is an AsyncIterable.
        gate = asyncio.Event()

        async def fragments():
            yield "⏳"
            await gate.wait()
            yield "⌛️"

        asyncio.create_task(self.connection.send(fragments()))
        await asyncio.sleep(0)  # let the event loop start the task
        await self.assertFrameSent(
            Frame(Opcode.TEXT, "⏳".encode(), fin=False),
        )

        asyncio.create_task(self.connection.send("✅"))
        await asyncio.sleep(0)  # let the event loop start the task
        await self.assertNoFrameSent()

        gate.set()
        await asyncio.sleep(0)  # run the event loop
        await asyncio.sleep(0)  # three times in order
        await asyncio.sleep(0)  # to send three frames
        await self.assertFramesSent(
            [
                Frame(Opcode.CONT, "⌛️".encode(), fin=False),
                Frame(Opcode.CONT, b"", fin=True),
                Frame(Opcode.TEXT, "✅".encode()),
            ]
        )

    async def test_send_while_send_blocked(self):
        """send waits for a blocked call to send to complete."""
        # This test fails if the guard with send_in_progress is removed
        # from send() in the case when message is an Iterable.
        self.connection.pause_writing()

        asyncio.create_task(self.connection.send(["⏳", "⌛️"]))
        await asyncio.sleep(0)  # let the event loop start the task
        await self.assertFrameSent(
            Frame(Opcode.TEXT, "⏳".encode(), fin=False),
        )

        asyncio.create_task(self.connection.send("✅"))
        await asyncio.sleep(0)  # let the event loop start the task
        await self.assertNoFrameSent()

        self.connection.resume_writing()
        await asyncio.sleep(0)  # run the event loop
        await asyncio.sleep(0)  # three times in order
        await asyncio.sleep(0)  # to send three frames
        await self.assertFramesSent(
            [
                Frame(Opcode.CONT, "⌛️".encode(), fin=False),
                Frame(Opcode.CONT, b"", fin=True),
                Frame(Opcode.TEXT, "✅".encode()),
            ]
        )

    async def test_send_while_send_async_blocked(self):
        """send waits for a blocked call to send to complete."""
        # This test fails if the guard with send_in_progress is removed
        # from send() in the case when message is an AsyncIterable.
        self.connection.pause_writing()

        async def fragments():
            yield "⏳"
            yield "⌛️"

        asyncio.create_task(self.connection.send(fragments()))
        await asyncio.sleep(0)  # let the event loop start the task
        await self.assertFrameSent(
            Frame(Opcode.TEXT, "⏳".encode(), fin=False),
        )

        asyncio.create_task(self.connection.send("✅"))
        await asyncio.sleep(0)  # let the event loop start the task
        await self.assertNoFrameSent()

        self.connection.resume_writing()
        await asyncio.sleep(0)  # run the event loop
        await asyncio.sleep(0)  # three times in order
        await asyncio.sleep(0)  # to send three frames
        await self.assertFramesSent(
            [
                Frame(Opcode.CONT, "⌛️".encode(), fin=False),
                Frame(Opcode.CONT, b"", fin=True),
                Frame(Opcode.TEXT, "✅".encode()),
            ]
        )

    async def test_send_empty_iterable(self):
        """send does nothing when called with an empty iterable."""
        await self.connection.send([])
        await self.connection.close()
        self.assertEqual(await alist(self.remote_connection), [])

    async def test_send_mixed_iterable(self):
        """send raises TypeError when called with an iterable of inconsistent types."""
        with self.assertRaises(TypeError):
            await self.connection.send(["😀", b"\xfe\xff"])

    async def test_send_unsupported_iterable(self):
        """send raises TypeError when called with an iterable of unsupported type."""
        with self.assertRaises(TypeError):
            await self.connection.send([None])

    async def test_send_empty_async_iterable(self):
        """send does nothing when called with an empty async iterable."""

        async def fragments():
            return
            yield  # pragma: no cover

        await self.connection.send(fragments())
        await self.connection.close()
        self.assertEqual(await alist(self.remote_connection), [])

    async def test_send_mixed_async_iterable(self):
        """send raises TypeError when called with an iterable of inconsistent types."""

        async def fragments():
            yield "😀"
            yield b"\xfe\xff"

        iterator = fragments()
        async with contextlib.aclosing(iterator):
            with self.assertRaises(TypeError):
                await self.connection.send(iterator)

    async def test_send_unsupported_async_iterable(self):
        """send raises TypeError when called with an iterable of unsupported type."""

        async def fragments():
            yield None

        iterator = fragments()
        async with contextlib.aclosing(iterator):
            with self.assertRaises(TypeError):
                await self.connection.send(iterator)

    async def test_send_dict(self):
        """send raises TypeError when called with a dict."""
        with self.assertRaises(TypeError):
            await self.connection.send({"type": "object"})

    async def test_send_unsupported_type(self):
        """send raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            await self.connection.send(None)

    # Test close.

    async def test_close(self):
        """close sends a close frame."""
        await self.connection.close()
        await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

    async def test_close_explicit_code_reason(self):
        """close sends a close frame with a given code and reason."""
        await self.connection.close(CloseCode.GOING_AWAY, "bye!")
        await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe9bye!"))

    async def test_close_waits_for_close_frame(self):
        """close waits for a close frame then EOF before returning."""
        t0 = self.loop.time()
        async with self.delay_frames_rcvd(MS), self.delay_eof_rcvd(MS):
            await self.connection.close()
        t1 = self.loop.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    async def test_close_waits_for_connection_closed(self):
        """close waits for EOF before returning."""
        if self.LOCAL is SERVER:
            self.skipTest("only relevant on the client-side")

        t0 = self.loop.time()
        async with self.delay_eof_rcvd(MS):
            await self.connection.close()
        t1 = self.loop.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    async def test_close_no_timeout_waits_for_close_frame(self):
        """close without timeout waits for a close frame then EOF before returning."""
        self.connection.close_timeout = None

        t0 = self.loop.time()
        async with self.delay_frames_rcvd(MS), self.delay_eof_rcvd(MS):
            await self.connection.close()
        t1 = self.loop.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    async def test_close_no_timeout_waits_for_connection_closed(self):
        """close without timeout waits for EOF before returning."""
        if self.LOCAL is SERVER:
            self.skipTest("only relevant on the client-side")

        self.connection.close_timeout = None

        t0 = self.loop.time()
        async with self.delay_eof_rcvd(MS):
            await self.connection.close()
        t1 = self.loop.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    async def test_close_timeout_waiting_for_close_frame(self):
        """close times out if no close frame is received."""
        t0 = self.loop.time()
        async with self.drop_eof_rcvd(), self.drop_frames_rcvd():
            await self.connection.close()
        t1 = self.loop.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.ABNORMAL_CLOSURE)
        self.assertGreater(t1 - t0, 2 * MS)

        with self.assertRaises(ConnectionClosedError) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); no close frame received")
        self.assertIsInstance(exc.__cause__, TimeoutError)

    async def test_close_timeout_waiting_for_connection_closed(self):
        """close times out if EOF isn't received."""
        if self.LOCAL is SERVER:
            self.skipTest("only relevant on the client-side")

        t0 = self.loop.time()
        async with self.drop_eof_rcvd():
            await self.connection.close()
        t1 = self.loop.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, 2 * MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsInstance(exc.__cause__, TimeoutError)

    async def test_close_preserves_queued_messages(self):
        """close preserves messages buffered in the assembler."""
        await self.remote_connection.send("😀")
        await self.connection.close()

        self.assertEqual(await self.connection.recv(), "😀")
        with self.assertRaises(ConnectionClosedOK):
            await self.connection.recv()

    async def test_close_idempotency(self):
        """close does nothing if the connection is already closed."""
        await self.connection.close()
        await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

        await self.connection.close()
        await self.assertNoFrameSent()

    async def test_close_during_recv(self):
        """close aborts recv when called concurrently with recv."""

        async def closer():
            await asyncio.sleep(MS)
            await self.connection.close()

        asyncio.create_task(closer())
        with self.assertRaises(ConnectionClosedOK) as raised:
            await self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    async def test_close_during_recv_streaming(self):
        """close aborts recv_streaming when called concurrently with recv_streaming."""

        async def closer():
            await asyncio.sleep(MS)
            await self.connection.close()

        asyncio.create_task(closer())
        with self.assertRaises(ConnectionClosedOK) as raised:
            async for _ in self.connection.recv_streaming():
                self.fail("did not raise")

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    async def test_close_during_send(self):
        """close fails the connection when called concurrently with send."""
        close_gate = asyncio.Event()
        exit_gate = asyncio.Event()

        async def closer():
            await close_gate.wait()
            await self.connection.close()
            exit_gate.set()

        async def fragments():
            yield "⏳"
            close_gate.set()
            await exit_gate.wait()
            yield "⌛️"

        asyncio.create_task(closer())
        iterator = fragments()
        async with contextlib.aclosing(iterator):
            with self.assertRaises(ConnectionClosedError) as raised:
                await self.connection.send(iterator)

        exc = raised.exception
        self.assertEqual(
            str(exc),
            "sent 1011 (internal error) close during fragmented message; "
            "no close frame received",
        )
        self.assertIsNone(exc.__cause__)

    # Test wait_closed.

    async def test_wait_closed(self):
        """wait_closed waits for the connection to close."""
        wait_closed_task = asyncio.create_task(self.connection.wait_closed())
        await asyncio.sleep(0)  # let the event loop start wait_closed_task
        self.assertFalse(wait_closed_task.done())
        await self.connection.close()
        self.assertTrue(wait_closed_task.done())

    # Test ping.

    @patch("random.getrandbits")
    async def test_ping(self, getrandbits):
        """ping sends a ping frame with a random payload."""
        getrandbits.side_effect = itertools.count(1918987876)
        await self.connection.ping()
        getrandbits.assert_called_once_with(32)
        await self.assertFrameSent(Frame(Opcode.PING, b"rand"))

    async def test_ping_explicit_text(self):
        """ping sends a ping frame with a payload provided as text."""
        await self.connection.ping("ping")
        await self.assertFrameSent(Frame(Opcode.PING, b"ping"))

    async def test_ping_explicit_binary(self):
        """ping sends a ping frame with a payload provided as binary."""
        await self.connection.ping(b"ping")
        await self.assertFrameSent(Frame(Opcode.PING, b"ping"))

    async def test_acknowledge_ping(self):
        """ping is acknowledged by a pong with the same payload."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("this")
        await self.remote_connection.pong("this")
        async with asyncio_timeout(MS):
            await pong_received

    async def test_acknowledge_canceled_ping(self):
        """ping is acknowledged by a pong with the same payload after being canceled."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("this")
        pong_received.cancel()
        await self.remote_connection.pong("this")
        with self.assertRaises(asyncio.CancelledError):
            await pong_received

    async def test_acknowledge_ping_non_matching_pong(self):
        """ping isn't acknowledged by a pong with a different payload."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("this")
        await self.remote_connection.pong("that")
        with self.assertRaises(TimeoutError):
            async with asyncio_timeout(MS):
                await pong_received

    async def test_acknowledge_previous_ping(self):
        """ping is acknowledged by a pong for a later ping."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("this")
            await self.connection.ping("that")
        await self.remote_connection.pong("that")
        async with asyncio_timeout(MS):
            await pong_received

    async def test_acknowledge_previous_canceled_ping(self):
        """ping is acknowledged by a pong for a later ping after being canceled."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("this")
            pong_received_2 = await self.connection.ping("that")
        pong_received.cancel()
        await self.remote_connection.pong("that")
        async with asyncio_timeout(MS):
            await pong_received_2
        with self.assertRaises(asyncio.CancelledError):
            await pong_received

    async def test_terminate_ping_on_close(self):
        """ping is done with an exception when the connection is closed."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("this")
        await self.connection.close()
        with self.assertRaises(ConnectionClosedOK):
            await pong_received

    async def test_ping_duplicate_payload(self):
        """ping rejects the same payload until receiving the pong."""
        async with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = await self.connection.ping("idem")

        with self.assertRaises(ConcurrencyError) as raised:
            await self.connection.ping("idem")
        self.assertEqual(
            str(raised.exception),
            "already waiting for a pong with the same data",
        )

        await self.remote_connection.pong("idem")
        async with asyncio_timeout(MS):
            await pong_received

        await self.connection.ping("idem")  # doesn't raise an exception

    async def test_ping_unsupported_type(self):
        """ping raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            await self.connection.ping([])

    # Test pong.

    async def test_pong(self):
        """pong sends a pong frame."""
        await self.connection.pong()
        await self.assertFrameSent(Frame(Opcode.PONG, b""))

    async def test_pong_explicit_text(self):
        """pong sends a pong frame with a payload provided as text."""
        await self.connection.pong("pong")
        await self.assertFrameSent(Frame(Opcode.PONG, b"pong"))

    async def test_pong_explicit_binary(self):
        """pong sends a pong frame with a payload provided as binary."""
        await self.connection.pong(b"pong")
        await self.assertFrameSent(Frame(Opcode.PONG, b"pong"))

    async def test_pong_unsupported_type(self):
        """pong raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            await self.connection.pong([])

    # Test keepalive.

    @patch("random.getrandbits")
    async def test_keepalive(self, getrandbits):
        """keepalive sends pings at ping_interval and measures latency."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping_interval = 3 * MS
        self.connection.start_keepalive()
        self.assertIsNotNone(self.connection.keepalive_task)
        self.assertEqual(self.connection.latency, 0)
        # 3 ms: keepalive() sends a ping frame.
        # 3.x ms: a pong frame is received.
        await asyncio.sleep(4 * MS)
        # 4 ms: check that the ping frame was sent.
        await self.assertFrameSent(Frame(Opcode.PING, b"rand"))
        self.assertGreater(self.connection.latency, 0)
        self.assertLess(self.connection.latency, MS)

    async def test_disable_keepalive(self):
        """keepalive is disabled when ping_interval is None."""
        self.connection.ping_interval = None
        self.connection.start_keepalive()
        self.assertIsNone(self.connection.keepalive_task)

    @patch("random.getrandbits")
    async def test_keepalive_times_out(self, getrandbits):
        """keepalive closes the connection if ping_timeout elapses."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping_interval = 4 * MS
        self.connection.ping_timeout = 2 * MS
        async with self.drop_frames_rcvd():
            self.connection.start_keepalive()
            # 4 ms: keepalive() sends a ping frame.
            # 4.x ms: a pong frame is dropped.
            await asyncio.sleep(5 * MS)
        # 6 ms: no pong frame is received; the connection is closed.
        await asyncio.sleep(3 * MS)
        # 8 ms: check that the connection is closed.
        self.assertEqual(self.connection.state, State.CLOSED)

    @patch("random.getrandbits")
    async def test_keepalive_ignores_timeout(self, getrandbits):
        """keepalive ignores timeouts if ping_timeout isn't set."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping_interval = 4 * MS
        self.connection.ping_timeout = None
        async with self.drop_frames_rcvd():
            self.connection.start_keepalive()
            # 4 ms: keepalive() sends a ping frame.
            # 4.x ms: a pong frame is dropped.
            await asyncio.sleep(5 * MS)
        # 6 ms: no pong frame is received; the connection remains open.
        await asyncio.sleep(3 * MS)
        # 8 ms: check that the connection is still open.
        self.assertEqual(self.connection.state, State.OPEN)

    async def test_keepalive_terminates_while_sleeping(self):
        """keepalive task terminates while waiting to send a ping."""
        self.connection.ping_interval = 3 * MS
        self.connection.start_keepalive()
        await asyncio.sleep(MS)
        self.assertFalse(self.connection.keepalive_task.done())
        await self.connection.close()
        self.assertTrue(self.connection.keepalive_task.done())

    # test_keepalive_terminates_when_sending_ping_fails is not implemented
    # because sending a ping cannot fail in the asyncio implementation.

    async def test_keepalive_terminates_while_waiting_for_pong(self):
        """keepalive task terminates while waiting to receive a pong."""
        self.connection.ping_interval = MS
        self.connection.ping_timeout = 4 * MS
        async with self.drop_frames_rcvd():
            self.connection.start_keepalive()
            # 1 ms: keepalive() sends a ping frame.
            # 1.x ms: a pong frame is dropped.
            await asyncio.sleep(2 * MS)
        # 2 ms: close the connection before ping_timeout elapses.
        await self.connection.close()
        self.assertTrue(self.connection.keepalive_task.done())

    async def test_keepalive_reports_errors(self):
        """keepalive reports unexpected errors in logs."""
        self.connection.ping_interval = 2 * MS
        self.connection.start_keepalive()
        async with self.drop_frames_rcvd():
            # 2 ms: keepalive() sends a ping frame.
            # 2.x ms: a pong frame is dropped.
            await asyncio.sleep(3 * MS)
        # 3 ms: inject a fault: raise an exception in the pending pong waiter.
        pong_received = next(iter(self.connection.pending_pings.values()))[0]
        with self.assertLogs("websockets", logging.ERROR) as logs:
            pong_received.set_exception(Exception("BOOM"))
            await asyncio.sleep(0)
        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["keepalive ping failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["BOOM"],
        )

    # Test parameters.

    async def test_close_timeout(self):
        """close_timeout parameter configures close timeout."""
        connection = Connection(
            Protocol(self.LOCAL),
            close_timeout=42 * MS,
        )
        self.assertEqual(connection.close_timeout, 42 * MS)

    async def test_max_queue(self):
        """max_queue configures high-water mark of frames buffer."""
        connection = Connection(
            Protocol(self.LOCAL),
            max_queue=4,
        )
        connection.connection_made(Mock(spec=asyncio.Transport))
        self.assertEqual(connection.recv_messages.high, 4)

    async def test_max_queue_none(self):
        """max_queue disables high-water mark of frames buffer."""
        connection = Connection(
            Protocol(self.LOCAL),
            max_queue=None,
        )
        connection.connection_made(Mock(spec=asyncio.Transport))
        self.assertEqual(connection.recv_messages.high, None)
        self.assertEqual(connection.recv_messages.low, None)

    async def test_max_queue_tuple(self):
        """max_queue configures high-water and low-water marks of frames buffer."""
        connection = Connection(
            Protocol(self.LOCAL),
            max_queue=(4, 2),
        )
        connection.connection_made(Mock(spec=asyncio.Transport))
        self.assertEqual(connection.recv_messages.high, 4)
        self.assertEqual(connection.recv_messages.low, 2)

    async def test_write_limit(self):
        """write_limit parameter configures high-water mark of write buffer."""
        connection = Connection(
            Protocol(self.LOCAL),
            write_limit=4096,
        )
        transport = Mock(spec=asyncio.Transport)
        connection.connection_made(transport)
        transport.set_write_buffer_limits.assert_called_once_with(4096, None)

    async def test_write_limits(self):
        """write_limit parameter configures high and low-water marks of write buffer."""
        connection = Connection(
            Protocol(self.LOCAL),
            write_limit=(4096, 2048),
        )
        transport = Mock(spec=asyncio.Transport)
        connection.connection_made(transport)
        transport.set_write_buffer_limits.assert_called_once_with(4096, 2048)

    # Test attributes.

    async def test_id(self):
        """Connection has an id attribute."""
        self.assertIsInstance(self.connection.id, uuid.UUID)

    async def test_logger(self):
        """Connection has a logger attribute."""
        self.assertIsInstance(self.connection.logger, logging.LoggerAdapter)

    @patch("asyncio.Transport.get_extra_info")
    async def test_local_address(self, get_extra_info):
        """Connection has a local_address attribute."""
        get_extra_info.return_value = ("sock", 1234)
        self.assertEqual(self.connection.local_address, ("sock", 1234))
        get_extra_info.assert_called_with("sockname")

    @patch("asyncio.Transport.get_extra_info")
    async def test_remote_address(self, get_extra_info):
        """Connection has a remote_address attribute."""
        get_extra_info.return_value = ("peer", 1234)
        self.assertEqual(self.connection.remote_address, ("peer", 1234))
        get_extra_info.assert_called_with("peername")

    async def test_state(self):
        """Connection has a state attribute."""
        self.assertIs(self.connection.state, State.OPEN)

    async def test_request(self):
        """Connection has a request attribute."""
        self.assertIsNone(self.connection.request)

    async def test_response(self):
        """Connection has a response attribute."""
        self.assertIsNone(self.connection.response)

    async def test_subprotocol(self):
        """Connection has a subprotocol attribute."""
        self.assertIsNone(self.connection.subprotocol)

    async def test_close_code(self):
        """Connection has a close_code attribute."""
        self.assertIsNone(self.connection.close_code)

    async def test_close_reason(self):
        """Connection has a close_reason attribute."""
        self.assertIsNone(self.connection.close_reason)

    # Test reporting of network errors.

    async def test_writing_in_data_received_fails(self):
        """Error when responding to incoming frames is correctly reported."""
        # Inject a fault by shutting down the transport for writing — but not by
        # closing it because that would terminate the connection.
        self.transport.write_eof()
        # Receive a ping. Responding with a pong will fail.
        await self.remote_connection.ping()
        with self.assertRaises(ConnectionClosedError) as raised:
            await self.connection.recv()
        self.assertIsInstance(raised.exception.__cause__, RuntimeError)

    async def test_writing_in_send_context_fails(self):
        """Error when sending outgoing frame is correctly reported."""
        # Inject a fault by shutting down the transport for writing — but not by
        # closing it because that would terminate the connection.
        self.transport.write_eof()
        # Sending a pong will fail.
        with self.assertRaises(ConnectionClosedError) as raised:
            await self.connection.pong()
        self.assertIsInstance(raised.exception.__cause__, RuntimeError)

    # Test safety nets — catching all exceptions in case of bugs.

    @patch("websockets.protocol.Protocol.events_received", side_effect=AssertionError)
    async def test_unexpected_failure_in_data_received(self, events_received):
        """Unexpected internal error in data_received() is correctly reported."""
        await self.remote_connection.send("😀")
        # Reading the message will trigger the injected fault.
        with self.assertRaises(ConnectionClosedError) as raised:
            await self.connection.recv()
        self.assertIsInstance(raised.exception.__cause__, AssertionError)

    @patch("websockets.protocol.Protocol.send_text", side_effect=AssertionError)
    async def test_unexpected_failure_in_send_context(self, send_text):
        """Unexpected internal error in send_context() is correctly reported."""
        # Sending a message will trigger the injected fault.
        with self.assertRaises(ConnectionClosedError) as raised:
            await self.connection.send("😀")
        self.assertIsInstance(raised.exception.__cause__, AssertionError)

    # Test broadcast.

    async def test_broadcast_text(self):
        """broadcast broadcasts a text message."""
        broadcast([self.connection], "😀")
        await self.assertFrameSent(Frame(Opcode.TEXT, "😀".encode()))

    @unittest.skipIf(
        sys.version_info[:2] < (3, 11),
        "raise_exceptions requires Python 3.11+",
    )
    async def test_broadcast_text_reports_no_errors(self):
        """broadcast broadcasts a text message without raising exceptions."""
        broadcast([self.connection], "😀", raise_exceptions=True)
        await self.assertFrameSent(Frame(Opcode.TEXT, "😀".encode()))

    async def test_broadcast_binary(self):
        """broadcast broadcasts a binary message."""
        broadcast([self.connection], b"\x01\x02\xfe\xff")
        await self.assertFrameSent(Frame(Opcode.BINARY, b"\x01\x02\xfe\xff"))

    @unittest.skipIf(
        sys.version_info[:2] < (3, 11),
        "raise_exceptions requires Python 3.11+",
    )
    async def test_broadcast_binary_reports_no_errors(self):
        """broadcast broadcasts a binary message without raising exceptions."""
        broadcast([self.connection], b"\x01\x02\xfe\xff", raise_exceptions=True)
        await self.assertFrameSent(Frame(Opcode.BINARY, b"\x01\x02\xfe\xff"))

    async def test_broadcast_no_clients(self):
        """broadcast does nothing when called with an empty list of clients."""
        broadcast([], "😀")
        await self.assertNoFrameSent()

    async def test_broadcast_two_clients(self):
        """broadcast broadcasts a message to several clients."""
        broadcast([self.connection, self.connection], "😀")
        await self.assertFramesSent(
            [
                Frame(Opcode.TEXT, "😀".encode()),
                Frame(Opcode.TEXT, "😀".encode()),
            ]
        )

    async def test_broadcast_skips_closed_connection(self):
        """broadcast ignores closed connections."""
        await self.connection.close()
        await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

        with self.assertNoLogs("websockets", logging.WARNING):
            broadcast([self.connection], "😀")
        await self.assertNoFrameSent()

    async def test_broadcast_skips_closing_connection(self):
        """broadcast ignores closing connections."""
        async with self.delay_frames_rcvd(MS):
            close_task = asyncio.create_task(self.connection.close())
            await asyncio.sleep(0)
            await self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

            with self.assertNoLogs("websockets", logging.WARNING):
                broadcast([self.connection], "😀")
            await self.assertNoFrameSent()

        await close_task

    async def test_broadcast_skips_connection_with_send_blocked(self):
        """broadcast logs a warning when a connection is blocked in send."""
        gate = asyncio.Event()

        async def fragments():
            yield "⏳"
            await gate.wait()

        send_task = asyncio.create_task(self.connection.send(fragments()))
        await asyncio.sleep(MS)
        await self.assertFrameSent(Frame(Opcode.TEXT, "⏳".encode(), fin=False))

        with self.assertLogs("websockets", logging.WARNING) as logs:
            broadcast([self.connection], "😀")

        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["skipped broadcast: sending a fragmented message"],
        )

        gate.set()
        await send_task

    @unittest.skipIf(
        sys.version_info[:2] < (3, 11),
        "raise_exceptions requires Python 3.11+",
    )
    async def test_broadcast_reports_connection_with_send_blocked(self):
        """broadcast raises exceptions for connections blocked in send."""
        gate = asyncio.Event()

        async def fragments():
            yield "⏳"
            await gate.wait()

        send_task = asyncio.create_task(self.connection.send(fragments()))
        await asyncio.sleep(MS)
        await self.assertFrameSent(Frame(Opcode.TEXT, "⏳".encode(), fin=False))

        with self.assertRaises(ExceptionGroup) as raised:
            broadcast([self.connection], "😀", raise_exceptions=True)

        self.assertEqual(str(raised.exception), "skipped broadcast (1 sub-exception)")
        exc = raised.exception.exceptions[0]
        self.assertEqual(str(exc), "sending a fragmented message")
        self.assertIsInstance(exc, ConcurrencyError)

        gate.set()
        await send_task

    async def test_broadcast_skips_connection_failing_to_send(self):
        """broadcast logs a warning when a connection fails to send."""
        # Inject a fault by shutting down the transport for writing.
        self.transport.write_eof()

        with self.assertLogs("websockets", logging.WARNING) as logs:
            broadcast([self.connection], "😀")

        self.assertEqual(
            [record.getMessage() for record in logs.records],
            [
                "skipped broadcast: failed to write message: "
                "RuntimeError: Cannot call write() after write_eof()"
            ],
        )

    @unittest.skipIf(
        sys.version_info[:2] < (3, 11),
        "raise_exceptions requires Python 3.11+",
    )
    async def test_broadcast_reports_connection_failing_to_send(self):
        """broadcast raises exceptions for connections failing to send."""
        # Inject a fault by shutting down the transport for writing.
        self.transport.write_eof()

        with self.assertRaises(ExceptionGroup) as raised:
            broadcast([self.connection], "😀", raise_exceptions=True)

        self.assertEqual(str(raised.exception), "skipped broadcast (1 sub-exception)")
        exc = raised.exception.exceptions[0]
        self.assertEqual(str(exc), "failed to write message")
        self.assertIsInstance(exc, RuntimeError)
        cause = exc.__cause__
        self.assertEqual(str(cause), "Cannot call write() after write_eof()")
        self.assertIsInstance(cause, RuntimeError)

    async def test_broadcast_type_error(self):
        """broadcast raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            broadcast([self.connection], ["⏳", "⌛️"])


class ServerConnectionTests(ClientConnectionTests):
    LOCAL = SERVER
    REMOTE = CLIENT



from __future__ import annotations

import argparse
import asyncio
import os
import sys
from typing import Generator

from .asyncio.client import ClientConnection, connect
from .asyncio.messages import SimpleQueue
from .exceptions import ConnectionClosed
from .frames import Close
from .streams import StreamReader
from .version import version as websockets_version


__all__ = ["main"]


def print_during_input(string: str) -> None:
    sys.stdout.write(
        # Save cursor position
        "\N{ESC}7"
        # Add a new line
        "\N{LINE FEED}"
        # Move cursor up
        "\N{ESC}[A"
        # Insert blank line, scroll last line down
        "\N{ESC}[L"
        # Print string in the inserted blank line
        f"{string}\N{LINE FEED}"
        # Restore cursor position
        "\N{ESC}8"
        # Move cursor down
        "\N{ESC}[B"
    )
    sys.stdout.flush()


def print_over_input(string: str) -> None:
    sys.stdout.write(
        # Move cursor to beginning of line
        "\N{CARRIAGE RETURN}"
        # Delete current line
        "\N{ESC}[K"
        # Print string
        f"{string}\N{LINE FEED}"
    )
    sys.stdout.flush()


class ReadLines(asyncio.Protocol):
    def __init__(self) -> None:
        self.reader = StreamReader()
        self.messages: SimpleQueue[str] = SimpleQueue()

    def parse(self) -> Generator[None, None, None]:
        while True:
            sys.stdout.write("> ")
            sys.stdout.flush()
            line = yield from self.reader.read_line(sys.maxsize)
            self.messages.put(line.decode().rstrip("\r\n"))

    def connection_made(self, transport: asyncio.BaseTransport) -> None:
        self.parser = self.parse()
        next(self.parser)

    def data_received(self, data: bytes) -> None:
        self.reader.feed_data(data)
        next(self.parser)

    def eof_received(self) -> None:
        self.reader.feed_eof()
        # next(self.parser) isn't useful and would raise EOFError.

    def connection_lost(self, exc: Exception | None) -> None:
        self.reader.discard()
        self.messages.abort()


async def print_incoming_messages(websocket: ClientConnection) -> None:
    async for message in websocket:
        if isinstance(message, str):
            print_during_input("< " + message)
        else:
            print_during_input("< (binary) " + message.hex())


async def send_outgoing_messages(
    websocket: ClientConnection,
    messages: SimpleQueue[str],
) -> None:
    while True:
        try:
            message = await messages.get()
        except EOFError:
            break
        try:
            await websocket.send(message)
        except ConnectionClosed:  # pragma: no cover
            break


async def interactive_client(uri: str) -> None:
    try:
        websocket = await connect(uri)
    except Exception as exc:
        print(f"Failed to connect to {uri}: {exc}.")
        sys.exit(1)
    else:
        print(f"Connected to {uri}.")

    loop = asyncio.get_running_loop()
    transport, protocol = await loop.connect_read_pipe(ReadLines, sys.stdin)
    incoming = asyncio.create_task(
        print_incoming_messages(websocket),
    )
    outgoing = asyncio.create_task(
        send_outgoing_messages(websocket, protocol.messages),
    )
    try:
        await asyncio.wait(
            [incoming, outgoing],
            # Clean up and exit when the server closes the connection
            # or the user enters EOT (^D), whichever happens first.
            return_when=asyncio.FIRST_COMPLETED,
        )
    # asyncio.run() cancels the main task when the user triggers SIGINT (^C).
    # https://docs.python.org/3/library/asyncio-runner.html#handling-keyboard-interruption
    # Clean up and exit without re-raising CancelledError to prevent Python
    # from raising KeyboardInterrupt and displaying a stack track.
    except asyncio.CancelledError:  # pragma: no cover
        pass
    finally:
        incoming.cancel()
        outgoing.cancel()
        transport.close()

    await websocket.close()
    assert websocket.close_code is not None and websocket.close_reason is not None
    close_status = Close(websocket.close_code, websocket.close_reason)
    print_over_input(f"Connection closed: {close_status}.")


def main(argv: list[str] | None = None) -> None:
    parser = argparse.ArgumentParser(
        prog="websockets",
        description="Interactive WebSocket client.",
        add_help=False,
    )
    group = parser.add_mutually_exclusive_group()
    group.add_argument("--version", action="store_true")
    group.add_argument("uri", metavar="<uri>", nargs="?")
    args = parser.parse_args(argv)

    if args.version:
        print(f"websockets {websockets_version}")
        return

    if args.uri is None:
        parser.print_usage()
        sys.exit(2)

    # Enable VT100 to support ANSI escape codes in Command Prompt on Windows.
    # See https://github.com/python/cpython/issues/74261 for why this works.
    if sys.platform == "win32":
        os.system("")

    try:
        import readline  # noqa: F401
    except ImportError:  # readline isn't available on all platforms
        pass

    # Remove the try/except block when dropping Python < 3.11.
    try:
        asyncio.run(interactive_client(args.uri))
    except KeyboardInterrupt:  # pragma: no cover
        pass



import asyncio
import pathlib
import ssl
import threading
import warnings


try:
    # Ignore deprecation warnings raised by mitmproxy dependencies at import time.
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="passlib")
    warnings.filterwarnings("ignore", category=DeprecationWarning, module="pyasn1")

    from mitmproxy import ctx
    from mitmproxy.addons import core, next_layer, proxyauth, proxyserver, tlsconfig
    from mitmproxy.http import Response
    from mitmproxy.master import Master
    from mitmproxy.options import CONF_BASENAME, CONF_DIR, Options
except ImportError:
    pass


class RecordFlows:
    def __init__(self, on_running):
        self.running = on_running
        self.http_connects = []
        self.tcp_flows = []

    def http_connect(self, flow):
        self.http_connects.append(flow)

    def tcp_start(self, flow):
        self.tcp_flows.append(flow)

    def get_http_connects(self):
        http_connects, self.http_connects[:] = self.http_connects[:], []
        return http_connects

    def get_tcp_flows(self):
        tcp_flows, self.tcp_flows[:] = self.tcp_flows[:], []
        return tcp_flows

    def reset(self):
        self.http_connects = []
        self.tcp_flows = []


class AlterRequest:
    def load(self, loader):
        loader.add_option(
            name="break_http_connect",
            typespec=bool,
            default=False,
            help="Respond to HTTP CONNECT requests with a 999 status code.",
        )
        loader.add_option(
            name="close_http_connect",
            typespec=bool,
            default=False,
            help="Do not respond to HTTP CONNECT requests.",
        )

    def http_connect(self, flow):
        if ctx.options.break_http_connect:
            # mitmproxy can send a response with a status code not between 100
            # and 599, while websockets treats it as a protocol error.
            # This is used for testing HTTP parsing errors.
            flow.response = Response.make(999, "not a valid HTTP response")
        if ctx.options.close_http_connect:
            flow.kill()


class ProxyMixin:
    """
    Run mitmproxy in a background thread.

    While it's uncommon to run two event loops in two threads, tests for the
    asyncio implementation rely on this class too because it starts an event
    loop for mitm proxy once, then a new event loop for each test.
    """

    proxy_mode = None

    @classmethod
    async def run_proxy(cls):
        cls.proxy_loop = loop = asyncio.get_event_loop()
        cls.proxy_stop = stop = loop.create_future()

        cls.proxy_options = options = Options(
            mode=[cls.proxy_mode],
            # Don't intercept connections, but record them.
            ignore_hosts=["^localhost:", "^127.0.0.1:", "^::1:"],
            # This option requires mitmproxy 11.0.0, which requires Python 3.11.
            show_ignored_hosts=True,
        )
        cls.proxy_master = master = Master(options)
        master.addons.add(
            core.Core(),
            proxyauth.ProxyAuth(),
            proxyserver.Proxyserver(),
            next_layer.NextLayer(),
            tlsconfig.TlsConfig(),
            RecordFlows(on_running=cls.proxy_ready.set),
            AlterRequest(),
        )

        task = loop.create_task(cls.proxy_master.run())
        await stop

        for server in master.addons.get("proxyserver").servers:
            await server.stop()
        master.shutdown()
        await task

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # Ignore deprecation warnings raised by mitmproxy at run time.
        warnings.filterwarnings(
            "ignore", category=DeprecationWarning, module="mitmproxy"
        )

        cls.proxy_ready = threading.Event()
        cls.proxy_thread = threading.Thread(target=asyncio.run, args=(cls.run_proxy(),))
        cls.proxy_thread.start()
        cls.proxy_ready.wait()

        certificate = pathlib.Path(CONF_DIR) / f"{CONF_BASENAME}-ca-cert.pem"
        certificate = certificate.expanduser()
        cls.proxy_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
        cls.proxy_context.load_verify_locations(bytes(certificate))

    def get_http_connects(self):
        return self.proxy_master.addons.get("recordflows").get_http_connects()

    def get_tcp_flows(self):
        return self.proxy_master.addons.get("recordflows").get_tcp_flows()

    def assertNumFlows(self, num_tcp_flows):
        self.assertEqual(len(self.get_tcp_flows()), num_tcp_flows)

    def tearDown(self):
        record_tcp_flows = self.proxy_master.addons.get("recordflows")
        record_tcp_flows.reset()
        super().tearDown()

    @classmethod
    def tearDownClass(cls):
        cls.proxy_loop.call_soon_threadsafe(cls.proxy_stop.set_result, None)
        cls.proxy_thread.join()
        super().tearDownClass()



import asyncio
import dataclasses
import hmac
import http
import logging
import socket
import unittest

from websockets.asyncio.client import connect, unix_connect
from websockets.asyncio.compatibility import TimeoutError, asyncio_timeout
from websockets.asyncio.server import *
from websockets.exceptions import (
    ConnectionClosedError,
    ConnectionClosedOK,
    InvalidStatus,
    NegotiationError,
)
from websockets.http11 import Request, Response

from ..utils import (
    CLIENT_CONTEXT,
    MS,
    SERVER_CONTEXT,
    temp_unix_socket_path,
)
from .server import (
    EvalShellMixin,
    args,
    get_host_port,
    get_uri,
    handler,
)


class ServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    async def test_connection(self):
        """Server receives connection from client and the handshake succeeds."""
        async with serve(*args) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.protocol.state.name", "OPEN")

    async def test_connection_handler_returns(self):
        """Connection handler returns."""
        async with serve(*args) as server:
            async with connect(get_uri(server) + "/no-op") as client:
                with self.assertRaises(ConnectionClosedOK) as raised:
                    await client.recv()
                self.assertEqual(
                    str(raised.exception),
                    "received 1000 (OK); then sent 1000 (OK)",
                )

    async def test_connection_handler_raises_exception(self):
        """Connection handler raises an exception."""
        async with serve(*args) as server:
            async with connect(get_uri(server) + "/crash") as client:
                with self.assertRaises(ConnectionClosedError) as raised:
                    await client.recv()
                self.assertEqual(
                    str(raised.exception),
                    "received 1011 (internal error); then sent 1011 (internal error)",
                )

    async def test_existing_socket(self):
        """Server receives connection using a pre-existing socket."""
        with socket.create_server(("localhost", 0)) as sock:
            host, port = sock.getsockname()
            async with serve(handler, sock=sock):
                async with connect(f"ws://{host}:{port}/") as client:
                    await self.assertEval(client, "ws.protocol.state.name", "OPEN")

    async def test_select_subprotocol(self):
        """Server selects a subprotocol with the select_subprotocol callable."""

        def select_subprotocol(ws, subprotocols):
            ws.select_subprotocol_ran = True
            assert "chat" in subprotocols
            return "chat"

        async with serve(
            *args,
            subprotocols=["chat"],
            select_subprotocol=select_subprotocol,
        ) as server:
            async with connect(get_uri(server), subprotocols=["chat"]) as client:
                await self.assertEval(client, "ws.select_subprotocol_ran", "True")
                await self.assertEval(client, "ws.subprotocol", "chat")

    async def test_select_subprotocol_rejects_handshake(self):
        """Server rejects handshake if select_subprotocol raises NegotiationError."""

        def select_subprotocol(ws, subprotocols):
            raise NegotiationError

        async with serve(*args, select_subprotocol=select_subprotocol) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server)):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 400",
            )

    async def test_select_subprotocol_raises_exception(self):
        """Server returns an error if select_subprotocol raises an exception."""

        def select_subprotocol(ws, subprotocols):
            raise RuntimeError

        async with serve(*args, select_subprotocol=select_subprotocol) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server)):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 500",
            )

    async def test_compression_is_enabled(self):
        """Server enables compression by default."""
        async with serve(*args) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(
                    client,
                    "[type(ext).__name__ for ext in ws.protocol.extensions]",
                    "['PerMessageDeflate']",
                )

    async def test_disable_compression(self):
        """Server disables compression."""
        async with serve(*args, compression=None) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.protocol.extensions", "[]")

    async def test_process_request_returns_none(self):
        """Server runs process_request and continues the handshake."""

        def process_request(ws, request):
            self.assertIsInstance(request, Request)
            ws.process_request_ran = True

        async with serve(*args, process_request=process_request) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.process_request_ran", "True")

    async def test_async_process_request_returns_none(self):
        """Server runs async process_request and continues the handshake."""

        async def process_request(ws, request):
            self.assertIsInstance(request, Request)
            ws.process_request_ran = True

        async with serve(*args, process_request=process_request) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.process_request_ran", "True")

    async def test_process_request_returns_response(self):
        """Server aborts handshake if process_request returns a response."""

        def process_request(ws, request):
            return ws.respond(http.HTTPStatus.FORBIDDEN, "Forbidden")

        async def handler(ws):
            self.fail("handler must not run")

        with self.assertNoLogs("websockets", logging.ERROR):
            async with serve(
                handler, *args[1:], process_request=process_request
            ) as server:
                with self.assertRaises(InvalidStatus) as raised:
                    async with connect(get_uri(server)):
                        self.fail("did not raise")
                self.assertEqual(
                    str(raised.exception),
                    "server rejected WebSocket connection: HTTP 403",
                )

    async def test_async_process_request_returns_response(self):
        """Server aborts handshake if async process_request returns a response."""

        async def process_request(ws, request):
            return ws.respond(http.HTTPStatus.FORBIDDEN, "Forbidden")

        async def handler(ws):
            self.fail("handler must not run")

        with self.assertNoLogs("websockets", logging.ERROR):
            async with serve(
                handler, *args[1:], process_request=process_request
            ) as server:
                with self.assertRaises(InvalidStatus) as raised:
                    async with connect(get_uri(server)):
                        self.fail("did not raise")
                self.assertEqual(
                    str(raised.exception),
                    "server rejected WebSocket connection: HTTP 403",
                )

    async def test_process_request_raises_exception(self):
        """Server returns an error if process_request raises an exception."""

        def process_request(ws, request):
            raise RuntimeError("BOOM")

        with self.assertLogs("websockets", logging.ERROR) as logs:
            async with serve(*args, process_request=process_request) as server:
                with self.assertRaises(InvalidStatus) as raised:
                    async with connect(get_uri(server)):
                        self.fail("did not raise")
                self.assertEqual(
                    str(raised.exception),
                    "server rejected WebSocket connection: HTTP 500",
                )
        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["opening handshake failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["BOOM"],
        )

    async def test_async_process_request_raises_exception(self):
        """Server returns an error if async process_request raises an exception."""

        async def process_request(ws, request):
            raise RuntimeError("BOOM")

        with self.assertLogs("websockets", logging.ERROR) as logs:
            async with serve(*args, process_request=process_request) as server:
                with self.assertRaises(InvalidStatus) as raised:
                    async with connect(get_uri(server)):
                        self.fail("did not raise")
                self.assertEqual(
                    str(raised.exception),
                    "server rejected WebSocket connection: HTTP 500",
                )
        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["opening handshake failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["BOOM"],
        )

    async def test_process_response_returns_none(self):
        """Server runs process_response but keeps the handshake response."""

        def process_response(ws, request, response):
            self.assertIsInstance(request, Request)
            self.assertIsInstance(response, Response)
            ws.process_response_ran = True

        async with serve(*args, process_response=process_response) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.process_response_ran", "True")

    async def test_async_process_response_returns_none(self):
        """Server runs async process_response but keeps the handshake response."""

        async def process_response(ws, request, response):
            self.assertIsInstance(request, Request)
            self.assertIsInstance(response, Response)
            ws.process_response_ran = True

        async with serve(*args, process_response=process_response) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.process_response_ran", "True")

    async def test_process_response_modifies_response(self):
        """Server runs process_response and modifies the handshake response."""

        def process_response(ws, request, response):
            response.headers["X-ProcessResponse"] = "OK"

        async with serve(*args, process_response=process_response) as server:
            async with connect(get_uri(server)) as client:
                self.assertEqual(client.response.headers["X-ProcessResponse"], "OK")

    async def test_async_process_response_modifies_response(self):
        """Server runs async process_response and modifies the handshake response."""

        async def process_response(ws, request, response):
            response.headers["X-ProcessResponse"] = "OK"

        async with serve(*args, process_response=process_response) as server:
            async with connect(get_uri(server)) as client:
                self.assertEqual(client.response.headers["X-ProcessResponse"], "OK")

    async def test_process_response_replaces_response(self):
        """Server runs process_response and replaces the handshake response."""

        def process_response(ws, request, response):
            headers = response.headers.copy()
            headers["X-ProcessResponse"] = "OK"
            return dataclasses.replace(response, headers=headers)

        async with serve(*args, process_response=process_response) as server:
            async with connect(get_uri(server)) as client:
                self.assertEqual(client.response.headers["X-ProcessResponse"], "OK")

    async def test_async_process_response_replaces_response(self):
        """Server runs async process_response and replaces the handshake response."""

        async def process_response(ws, request, response):
            headers = response.headers.copy()
            headers["X-ProcessResponse"] = "OK"
            return dataclasses.replace(response, headers=headers)

        async with serve(*args, process_response=process_response) as server:
            async with connect(get_uri(server)) as client:
                self.assertEqual(client.response.headers["X-ProcessResponse"], "OK")

    async def test_process_response_raises_exception(self):
        """Server returns an error if process_response raises an exception."""

        def process_response(ws, request, response):
            raise RuntimeError("BOOM")

        with self.assertLogs("websockets", logging.ERROR) as logs:
            async with serve(*args, process_response=process_response) as server:
                with self.assertRaises(InvalidStatus) as raised:
                    async with connect(get_uri(server)):
                        self.fail("did not raise")
                self.assertEqual(
                    str(raised.exception),
                    "server rejected WebSocket connection: HTTP 500",
                )
        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["opening handshake failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["BOOM"],
        )

    async def test_async_process_response_raises_exception(self):
        """Server returns an error if async process_response raises an exception."""

        async def process_response(ws, request, response):
            raise RuntimeError("BOOM")

        with self.assertLogs("websockets", logging.ERROR) as logs:
            async with serve(*args, process_response=process_response) as server:
                with self.assertRaises(InvalidStatus) as raised:
                    async with connect(get_uri(server)):
                        self.fail("did not raise")
                self.assertEqual(
                    str(raised.exception),
                    "server rejected WebSocket connection: HTTP 500",
                )
        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["opening handshake failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["BOOM"],
        )

    async def test_override_server(self):
        """Server can override Server header with server_header."""
        async with serve(*args, server_header="Neo") as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.response.headers['Server']", "Neo")

    async def test_remove_server(self):
        """Server can remove Server header with server_header."""
        async with serve(*args, server_header=None) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(
                    client, "'Server' in ws.response.headers", "False"
                )

    async def test_keepalive_is_enabled(self):
        """Server enables keepalive and measures latency."""
        async with serve(*args, ping_interval=MS) as server:
            async with connect(get_uri(server)) as client:
                await client.send("ws.latency")
                latency = eval(await client.recv())
                self.assertEqual(latency, 0)
                await asyncio.sleep(2 * MS)
                await client.send("ws.latency")
                latency = eval(await client.recv())
                self.assertGreater(latency, 0)

    async def test_disable_keepalive(self):
        """Server disables keepalive."""
        async with serve(*args, ping_interval=None) as server:
            async with connect(get_uri(server)) as client:
                await asyncio.sleep(2 * MS)
                await client.send("ws.latency")
                latency = eval(await client.recv())
                self.assertEqual(latency, 0)

    async def test_logger(self):
        """Server accepts a logger argument."""
        logger = logging.getLogger("test")
        async with serve(*args, logger=logger) as server:
            self.assertEqual(server.logger.name, logger.name)

    async def test_custom_connection_factory(self):
        """Server runs ServerConnection factory provided in create_connection."""

        def create_connection(*args, **kwargs):
            server = ServerConnection(*args, **kwargs)
            server.create_connection_ran = True
            return server

        async with serve(*args, create_connection=create_connection) as server:
            async with connect(get_uri(server)) as client:
                await self.assertEval(client, "ws.create_connection_ran", "True")

    async def test_connections(self):
        """Server provides a connections property."""
        async with serve(*args) as server:
            self.assertEqual(server.connections, set())
            async with connect(get_uri(server)) as client:
                self.assertEqual(len(server.connections), 1)
                ws_id = str(next(iter(server.connections)).id)
                await self.assertEval(client, "ws.id", ws_id)
            self.assertEqual(server.connections, set())

    async def test_handshake_fails(self):
        """Server receives connection from client but the handshake fails."""

        def remove_key_header(self, request):
            del request.headers["Sec-WebSocket-Key"]

        async with serve(*args, process_request=remove_key_header) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server)):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 400",
            )

    async def test_timeout_during_handshake(self):
        """Server times out before receiving handshake request from client."""
        async with serve(*args, open_timeout=MS) as server:
            reader, writer = await asyncio.open_connection(*get_host_port(server))
            try:
                self.assertEqual(await reader.read(4096), b"")
            finally:
                writer.close()

    async def test_connection_closed_during_handshake(self):
        """Server reads EOF before receiving handshake request from client."""
        async with serve(*args) as server:
            _reader, writer = await asyncio.open_connection(*get_host_port(server))
            writer.close()

    async def test_junk_handshake(self):
        """Server closes the connection when receiving non-HTTP request from client."""
        with self.assertLogs("websockets", logging.ERROR) as logs:
            async with serve(*args) as server:
                reader, writer = await asyncio.open_connection(*get_host_port(server))
                writer.write(b"HELO relay.invalid\r\n")
                try:
                    # Wait for the server to close the connection.
                    self.assertEqual(await reader.read(4096), b"")
                finally:
                    writer.close()

        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["opening handshake failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["did not receive a valid HTTP request"],
        )
        self.assertEqual(
            [str(record.exc_info[1].__cause__) for record in logs.records],
            ["invalid HTTP request line: HELO relay.invalid"],
        )

    async def test_close_server_rejects_connecting_connections(self):
        """Server rejects connecting connections with HTTP 503 when closing."""

        async def process_request(ws, _request):
            while ws.server.is_serving():
                await asyncio.sleep(0)  # pragma: no cover

        async with serve(*args, process_request=process_request) as server:
            asyncio.get_running_loop().call_later(MS, server.close)
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server)):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 503",
            )

    async def test_close_server_closes_open_connections(self):
        """Server closes open connections with close code 1001 when closing."""
        async with serve(*args) as server:
            async with connect(get_uri(server)) as client:
                server.close()
                with self.assertRaises(ConnectionClosedOK) as raised:
                    await client.recv()
                self.assertEqual(
                    str(raised.exception),
                    "received 1001 (going away); then sent 1001 (going away)",
                )

    async def test_close_server_closes_open_connections_with_code_and_reason(self):
        """Server closes open connections with custom code and reason when closing."""
        async with serve(*args) as server:
            async with connect(get_uri(server)) as client:
                server.close(code=1012, reason="restarting")
                with self.assertRaises(ConnectionClosedError) as raised:
                    await client.recv()
                self.assertEqual(
                    str(raised.exception),
                    "received 1012 (service restart) restarting; "
                    "then sent 1012 (service restart) restarting",
                )

    async def test_close_server_keeps_connections_open(self):
        """Server waits for client to close open connections when closing."""
        async with serve(*args) as server:
            async with connect(get_uri(server)) as client:
                server.close(close_connections=False)

                # Server cannot receive new connections.
                await asyncio.sleep(0)
                self.assertFalse(server.sockets)

                # The server waits for the client to close the connection.
                with self.assertRaises(TimeoutError):
                    async with asyncio_timeout(MS):
                        await server.wait_closed()

                # Once the client closes the connection, the server terminates.
                await client.close()
                async with asyncio_timeout(MS):
                    await server.wait_closed()

    async def test_close_server_keeps_handlers_running(self):
        """Server waits for connection handlers to terminate."""
        async with serve(*args) as server:
            async with connect(get_uri(server) + "/delay") as client:
                # Delay termination of connection handler.
                await client.send(str(3 * MS))

                server.close()

                # The server waits for the connection handler to terminate.
                with self.assertRaises(TimeoutError):
                    async with asyncio_timeout(2 * MS):
                        await server.wait_closed()

                # Set a large timeout here, else the test becomes flaky.
                async with asyncio_timeout(5 * MS):
                    await server.wait_closed()


SSL_OBJECT = "ws.transport.get_extra_info('ssl_object')"


class SecureServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    async def test_connection(self):
        """Server receives secure connection from client."""
        async with serve(*args, ssl=SERVER_CONTEXT) as server:
            async with connect(get_uri(server), ssl=CLIENT_CONTEXT) as client:
                await self.assertEval(client, "ws.protocol.state.name", "OPEN")
                await self.assertEval(client, SSL_OBJECT + ".version()[:3]", "TLS")

    async def test_timeout_during_tls_handshake(self):
        """Server times out before receiving TLS handshake request from client."""
        async with serve(*args, ssl=SERVER_CONTEXT, open_timeout=MS) as server:
            reader, writer = await asyncio.open_connection(*get_host_port(server))
            try:
                self.assertEqual(await reader.read(4096), b"")
            finally:
                writer.close()

    async def test_connection_closed_during_tls_handshake(self):
        """Server reads EOF before receiving TLS handshake request from client."""
        async with serve(*args, ssl=SERVER_CONTEXT) as server:
            _reader, writer = await asyncio.open_connection(*get_host_port(server))
            writer.close()


@unittest.skipUnless(hasattr(socket, "AF_UNIX"), "this test requires Unix sockets")
class UnixServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    async def test_connection(self):
        """Server receives connection from client over a Unix socket."""
        with temp_unix_socket_path() as path:
            async with unix_serve(handler, path):
                async with unix_connect(path) as client:
                    await self.assertEval(client, "ws.protocol.state.name", "OPEN")


@unittest.skipUnless(hasattr(socket, "AF_UNIX"), "this test requires Unix sockets")
class SecureUnixServerTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    async def test_connection(self):
        """Server receives secure connection from client over a Unix socket."""
        with temp_unix_socket_path() as path:
            async with unix_serve(handler, path, ssl=SERVER_CONTEXT):
                async with unix_connect(path, ssl=CLIENT_CONTEXT) as client:
                    await self.assertEval(client, "ws.protocol.state.name", "OPEN")
                    await self.assertEval(client, SSL_OBJECT + ".version()[:3]", "TLS")


class ServerUsageErrorsTests(unittest.IsolatedAsyncioTestCase):
    async def test_unix_without_path_or_sock(self):
        """Unix server requires path when sock isn't provided."""
        with self.assertRaises(ValueError) as raised:
            await unix_serve(handler)
        self.assertEqual(
            str(raised.exception),
            "path was not specified, and no sock specified",
        )

    async def test_unix_with_path_and_sock(self):
        """Unix server rejects path when sock is provided."""
        sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.addCleanup(sock.close)
        with self.assertRaises(ValueError) as raised:
            await unix_serve(handler, path="/", sock=sock)
        self.assertEqual(
            str(raised.exception),
            "path and sock can not be specified at the same time",
        )

    async def test_invalid_subprotocol(self):
        """Server rejects single value of subprotocols."""
        with self.assertRaises(TypeError) as raised:
            await serve(*args, subprotocols="chat")
        self.assertEqual(
            str(raised.exception),
            "subprotocols must be a list, not a str",
        )

    async def test_unsupported_compression(self):
        """Server rejects incorrect value of compression."""
        with self.assertRaises(ValueError) as raised:
            await serve(*args, compression=False)
        self.assertEqual(
            str(raised.exception),
            "unsupported compression: False",
        )


class BasicAuthTests(EvalShellMixin, unittest.IsolatedAsyncioTestCase):
    async def test_valid_authorization(self):
        """basic_auth authenticates client with HTTP Basic Authentication."""
        async with serve(
            *args,
            process_request=basic_auth(credentials=("hello", "iloveyou")),
        ) as server:
            async with connect(
                get_uri(server),
                additional_headers={"Authorization": "Basic aGVsbG86aWxvdmV5b3U="},
            ) as client:
                await self.assertEval(client, "ws.username", "hello")

    async def test_missing_authorization(self):
        """basic_auth rejects client without credentials."""
        async with serve(
            *args,
            process_request=basic_auth(credentials=("hello", "iloveyou")),
        ) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(get_uri(server)):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 401",
            )

    async def test_unsupported_authorization(self):
        """basic_auth rejects client with unsupported credentials."""
        async with serve(
            *args,
            process_request=basic_auth(credentials=("hello", "iloveyou")),
        ) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(
                    get_uri(server),
                    additional_headers={"Authorization": "Negotiate ..."},
                ):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 401",
            )

    async def test_authorization_with_unknown_username(self):
        """basic_auth rejects client with unknown username."""
        async with serve(
            *args,
            process_request=basic_auth(credentials=("hello", "iloveyou")),
        ) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(
                    get_uri(server),
                    additional_headers={"Authorization": "Basic YnllOnlvdWxvdmVtZQ=="},
                ):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 401",
            )

    async def test_authorization_with_incorrect_password(self):
        """basic_auth rejects client with incorrect password."""
        async with serve(
            *args,
            process_request=basic_auth(credentials=("hello", "changeme")),
        ) as server:
            with self.assertRaises(InvalidStatus) as raised:
                async with connect(
                    get_uri(server),
                    additional_headers={"Authorization": "Basic aGVsbG86aWxvdmV5b3U="},
                ):
                    self.fail("did not raise")
            self.assertEqual(
                str(raised.exception),
                "server rejected WebSocket connection: HTTP 401",
            )

    async def test_list_of_credentials(self):
        """basic_auth accepts a list of hard coded credentials."""
        async with serve(
            *args,
            process_request=basic_auth(
                credentials=[
                    ("hello", "iloveyou"),
                    ("bye", "youloveme"),
                ]
            ),
        ) as server:
            async with connect(
                get_uri(server),
                additional_headers={"Authorization": "Basic YnllOnlvdWxvdmVtZQ=="},
            ) as client:
                await self.assertEval(client, "ws.username", "bye")

    async def test_check_credentials_function(self):
        """basic_auth accepts a check_credentials function."""

        def check_credentials(username, password):
            return hmac.compare_digest(password, "iloveyou")

        async with serve(
            *args,
            process_request=basic_auth(check_credentials=check_credentials),
        ) as server:
            async with connect(
                get_uri(server),
                additional_headers={"Authorization": "Basic aGVsbG86aWxvdmV5b3U="},
            ) as client:
                await self.assertEval(client, "ws.username", "hello")

    async def test_check_credentials_coroutine(self):
        """basic_auth accepts a check_credentials coroutine."""

        async def check_credentials(username, password):
            return hmac.compare_digest(password, "iloveyou")

        async with serve(
            *args,
            process_request=basic_auth(check_credentials=check_credentials),
        ) as server:
            async with connect(
                get_uri(server),
                additional_headers={"Authorization": "Basic aGVsbG86aWxvdmV5b3U="},
            ) as client:
                await self.assertEval(client, "ws.username", "hello")

    async def test_without_credentials_or_check_credentials(self):
        """basic_auth requires either credentials or check_credentials."""
        with self.assertRaises(ValueError) as raised:
            basic_auth()
        self.assertEqual(
            str(raised.exception),
            "provide either credentials or check_credentials",
        )

    async def test_with_credentials_and_check_credentials(self):
        """basic_auth requires only one of credentials and check_credentials."""
        with self.assertRaises(ValueError) as raised:
            basic_auth(
                credentials=("hello", "iloveyou"),
                check_credentials=lambda: False,  # pragma: no cover
            )
        self.assertEqual(
            str(raised.exception),
            "provide either credentials or check_credentials",
        )

    async def test_bad_credentials(self):
        """basic_auth receives an unsupported credentials argument."""
        with self.assertRaises(TypeError) as raised:
            basic_auth(credentials=42)
        self.assertEqual(
            str(raised.exception),
            "invalid credentials argument: 42",
        )

    async def test_bad_list_of_credentials(self):
        """basic_auth receives an unsupported credentials argument."""
        with self.assertRaises(TypeError) as raised:
            basic_auth(credentials=[42])
        self.assertEqual(
            str(raised.exception),
            "invalid credentials argument: [42]",
        )



import unittest

from websockets.datastructures import Headers
from websockets.legacy.exceptions import *


class ExceptionsTests(unittest.TestCase):
    def test_str(self):
        for exception, exception_str in [
            (
                InvalidStatusCode(403, Headers()),
                "server rejected WebSocket connection: HTTP 403",
            ),
            (
                AbortHandshake(200, Headers(), b"OK\n"),
                "HTTP 200, 0 headers, 3 bytes",
            ),
            (
                RedirectHandshake("wss://example.com"),
                "redirect to wss://example.com",
            ),
        ]:
            with self.subTest(exception=exception):
                self.assertEqual(str(exception), exception_str)



import contextlib
import itertools
import logging
import socket
import threading
import time
import uuid
from unittest.mock import Mock, patch

from websockets.exceptions import (
    ConcurrencyError,
    ConnectionClosedError,
    ConnectionClosedOK,
)
from websockets.frames import CloseCode, Frame, Opcode
from websockets.protocol import CLIENT, SERVER, Protocol, State
from websockets.sync.connection import *

from ..protocol import RecordingProtocol
from ..utils import MS
from .connection import InterceptingConnection
from .utils import ThreadTestCase


# Connection implements symmetrical behavior between clients and servers.
# All tests run on the client side and the server side to validate this.


class ClientConnectionTests(ThreadTestCase):
    LOCAL = CLIENT
    REMOTE = SERVER

    def setUp(self):
        socket_, remote_socket = socket.socketpair()
        protocol = Protocol(self.LOCAL)
        remote_protocol = RecordingProtocol(self.REMOTE)
        self.connection = Connection(socket_, protocol, close_timeout=2 * MS)
        self.remote_connection = InterceptingConnection(remote_socket, remote_protocol)

    def tearDown(self):
        self.remote_connection.close()
        self.connection.close()

    # Test helpers built upon RecordingProtocol and InterceptingConnection.

    def wait_for_remote_side(self):
        """Wait for the remote side to process messages."""
        # We don't have a way to tell if the remote side is blocked on I/O.
        # The sync tests still run faster than the asyncio and trio tests :-)
        time.sleep(MS)

    def assertFrameSent(self, frame):
        """Check that a single frame was sent."""
        self.wait_for_remote_side()
        self.assertEqual(self.remote_connection.protocol.get_frames_rcvd(), [frame])

    def assertNoFrameSent(self):
        """Check that no frame was sent."""
        self.wait_for_remote_side()
        self.assertEqual(self.remote_connection.protocol.get_frames_rcvd(), [])

    @contextlib.contextmanager
    def delay_frames_rcvd(self, delay):
        """Delay frames before they're received by the connection."""
        with self.remote_connection.delay_frames_sent(delay):
            yield
            self.wait_for_remote_side()

    @contextlib.contextmanager
    def delay_eof_rcvd(self, delay):
        """Delay EOF before it's received by the connection."""
        with self.remote_connection.delay_eof_sent(delay):
            yield
            self.wait_for_remote_side()

    @contextlib.contextmanager
    def drop_frames_rcvd(self):
        """Drop frames before they're received by the connection."""
        with self.remote_connection.drop_frames_sent():
            yield
            self.wait_for_remote_side()

    @contextlib.contextmanager
    def drop_eof_rcvd(self):
        """Drop EOF before it's received by the connection."""
        with self.remote_connection.drop_eof_sent():
            yield
            self.wait_for_remote_side()

    # Test __enter__ and __exit__.

    def test_enter(self):
        """__enter__ returns the connection itself."""
        with self.connection as connection:
            self.assertIs(connection, self.connection)

    def test_exit(self):
        """__exit__ closes the connection with code 1000."""
        with self.connection:
            self.assertNoFrameSent()
        self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

    def test_exit_with_exception(self):
        """__exit__ with an exception closes the connection with code 1011."""
        with self.assertRaises(RuntimeError):
            with self.connection:
                raise RuntimeError
        self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xf3"))

    # Test __iter__.

    def test_iter_text(self):
        """__iter__ yields text messages."""
        iterator = iter(self.connection)
        with contextlib.closing(iterator):
            self.remote_connection.send("😀")
            self.assertEqual(next(iterator), "😀")
            self.remote_connection.send("😀")
            self.assertEqual(next(iterator), "😀")

    def test_iter_binary(self):
        """__iter__ yields binary messages."""
        iterator = iter(self.connection)
        with contextlib.closing(iterator):
            self.remote_connection.send(b"\x01\x02\xfe\xff")
            self.assertEqual(next(iterator), b"\x01\x02\xfe\xff")
            self.remote_connection.send(b"\x01\x02\xfe\xff")
            self.assertEqual(next(iterator), b"\x01\x02\xfe\xff")

    def test_iter_mixed(self):
        """__iter__ yields a mix of text and binary messages."""
        iterator = iter(self.connection)
        with contextlib.closing(iterator):
            self.remote_connection.send("😀")
            self.assertEqual(next(iterator), "😀")
            self.remote_connection.send(b"\x01\x02\xfe\xff")
            self.assertEqual(next(iterator), b"\x01\x02\xfe\xff")

    def test_iter_connection_closed_ok(self):
        """__iter__ terminates after a normal closure."""
        iterator = iter(self.connection)
        with contextlib.closing(iterator):
            self.remote_connection.close()
            with self.assertRaises(StopIteration):
                next(iterator)

    def test_iter_connection_closed_error(self):
        """__iter__ raises ConnectionClosedError after an error."""
        iterator = iter(self.connection)
        with contextlib.closing(iterator):
            self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
            with self.assertRaises(ConnectionClosedError):
                next(iterator)

    # Test recv.

    def test_recv_text(self):
        """recv receives a text message."""
        self.remote_connection.send("😀")
        self.assertEqual(self.connection.recv(), "😀")

    def test_recv_binary(self):
        """recv receives a binary message."""
        self.remote_connection.send(b"\x01\x02\xfe\xff")
        self.assertEqual(self.connection.recv(), b"\x01\x02\xfe\xff")

    def test_recv_text_as_bytes(self):
        """recv receives a text message as bytes."""
        self.remote_connection.send("😀")
        self.assertEqual(self.connection.recv(decode=False), "😀".encode())

    def test_recv_binary_as_text(self):
        """recv receives a binary message as a str."""
        self.remote_connection.send("😀".encode())
        self.assertEqual(self.connection.recv(decode=True), "😀")

    def test_recv_fragmented_text(self):
        """recv receives a fragmented text message."""
        self.remote_connection.send(["😀", "😀"])
        self.assertEqual(self.connection.recv(), "😀😀")

    def test_recv_fragmented_binary(self):
        """recv receives a fragmented binary message."""
        self.remote_connection.send([b"\x01\x02", b"\xfe\xff"])
        self.assertEqual(self.connection.recv(), b"\x01\x02\xfe\xff")

    def test_recv_connection_closed_ok(self):
        """recv raises ConnectionClosedOK after a normal closure."""
        self.remote_connection.close()
        with self.assertRaises(ConnectionClosedOK):
            self.connection.recv()

    def test_recv_connection_closed_error(self):
        """recv raises ConnectionClosedError after an error."""
        self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
        with self.assertRaises(ConnectionClosedError):
            self.connection.recv()

    def test_recv_non_utf8_text(self):
        """recv receives a non-UTF-8 text message."""
        self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
        with self.assertRaises(ConnectionClosedError) as raised:
            self.connection.recv()
        self.assertEqual(raised.exception.sent.code, CloseCode.INVALID_DATA)

    def test_recv_during_recv(self):
        """recv raises ConcurrencyError when called concurrently."""
        with self.run_in_thread(self.connection.recv):
            try:
                with self.assertRaises(ConcurrencyError) as raised:
                    self.connection.recv()
            finally:
                self.remote_connection.send("")
        self.assertEqual(
            str(raised.exception),
            "cannot call recv while another thread "
            "is already running recv or recv_streaming",
        )

    def test_recv_during_recv_streaming(self):
        """recv raises ConcurrencyError when called concurrently with recv_streaming."""
        with self.run_in_thread(lambda: list(self.connection.recv_streaming())):
            try:
                with self.assertRaises(ConcurrencyError) as raised:
                    self.connection.recv()
            finally:
                self.remote_connection.send("")
        self.assertEqual(
            str(raised.exception),
            "cannot call recv while another thread "
            "is already running recv or recv_streaming",
        )

    # Test recv_streaming.

    def test_recv_streaming_text(self):
        """recv_streaming receives a text message."""
        self.remote_connection.send("😀")
        self.assertEqual(
            list(self.connection.recv_streaming()),
            ["😀"],
        )

    def test_recv_streaming_binary(self):
        """recv_streaming receives a binary message."""
        self.remote_connection.send(b"\x01\x02\xfe\xff")
        self.assertEqual(
            list(self.connection.recv_streaming()),
            [b"\x01\x02\xfe\xff"],
        )

    def test_recv_streaming_text_as_bytes(self):
        """recv_streaming receives a text message as bytes."""
        self.remote_connection.send("😀")
        self.assertEqual(
            list(self.connection.recv_streaming(decode=False)),
            ["😀".encode()],
        )

    def test_recv_streaming_binary_as_str(self):
        """recv_streaming receives a binary message as a str."""
        self.remote_connection.send("😀".encode())
        self.assertEqual(
            list(self.connection.recv_streaming(decode=True)),
            ["😀"],
        )

    def test_recv_streaming_fragmented_text(self):
        """recv_streaming receives a fragmented text message."""
        self.remote_connection.send(["😀", "😀"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            list(self.connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    def test_recv_streaming_fragmented_binary(self):
        """recv_streaming receives a fragmented binary message."""
        self.remote_connection.send([b"\x01\x02", b"\xfe\xff"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            list(self.connection.recv_streaming()),
            [b"\x01\x02", b"\xfe\xff", b""],
        )

    def test_recv_streaming_connection_closed_ok(self):
        """recv_streaming raises ConnectionClosedOK after a normal closure."""
        self.remote_connection.close()
        with self.assertRaises(ConnectionClosedOK):
            for _ in self.connection.recv_streaming():
                self.fail("did not raise")

    def test_recv_streaming_connection_closed_error(self):
        """recv_streaming raises ConnectionClosedError after an error."""
        self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
        with self.assertRaises(ConnectionClosedError):
            for _ in self.connection.recv_streaming():
                self.fail("did not raise")

    def test_recv_streaming_non_utf8_text(self):
        """recv_streaming receives a non-UTF-8 text message."""
        self.remote_connection.send(b"\x01\x02\xfe\xff", text=True)
        with self.assertRaises(ConnectionClosedError) as raised:
            list(self.connection.recv_streaming())
        self.assertEqual(raised.exception.sent.code, CloseCode.INVALID_DATA)

    def test_recv_streaming_during_recv(self):
        """recv_streaming raises ConcurrencyError when called concurrently with recv."""
        with self.run_in_thread(self.connection.recv):
            try:
                with self.assertRaises(ConcurrencyError) as raised:
                    for _ in self.connection.recv_streaming():
                        self.fail("did not raise")
            finally:
                self.remote_connection.send("")
        self.assertEqual(
            str(raised.exception),
            "cannot call recv_streaming while another thread "
            "is already running recv or recv_streaming",
        )

    def test_recv_streaming_during_recv_streaming(self):
        """recv_streaming raises ConcurrencyError when called concurrently."""
        with self.run_in_thread(lambda: list(self.connection.recv_streaming())):
            try:
                with self.assertRaises(ConcurrencyError) as raised:
                    for _ in self.connection.recv_streaming():
                        self.fail("did not raise")
            finally:
                self.remote_connection.send("")
        self.assertEqual(
            str(raised.exception),
            "cannot call recv_streaming while another thread "
            "is already running recv or recv_streaming",
        )

    # Test send.

    def test_send_text(self):
        """send sends a text message."""
        self.connection.send("😀")
        self.assertEqual(self.remote_connection.recv(), "😀")

    def test_send_binary(self):
        """send sends a binary message."""
        self.connection.send(b"\x01\x02\xfe\xff")
        self.assertEqual(self.remote_connection.recv(), b"\x01\x02\xfe\xff")

    def test_send_binary_from_str(self):
        """send sends a binary message from a str."""
        self.connection.send("😀", text=False)
        self.assertEqual(self.remote_connection.recv(), "😀".encode())

    def test_send_text_from_bytes(self):
        """send sends a text message from bytes."""
        self.connection.send("😀".encode(), text=True)
        self.assertEqual(self.remote_connection.recv(), "😀")

    def test_send_fragmented_text(self):
        """send sends a fragmented text message."""
        self.connection.send(["😀", "😀"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            list(self.remote_connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    def test_send_fragmented_binary(self):
        """send sends a fragmented binary message."""
        self.connection.send([b"\x01\x02", b"\xfe\xff"])
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            list(self.remote_connection.recv_streaming()),
            [b"\x01\x02", b"\xfe\xff", b""],
        )

    def test_send_fragmented_binary_from_str(self):
        """send sends a fragmented binary message from a str."""
        self.connection.send(["😀", "😀"], text=False)
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            list(self.remote_connection.recv_streaming()),
            ["😀".encode(), "😀".encode(), b""],
        )

    def test_send_fragmented_text_from_bytes(self):
        """send sends a fragmented text message from bytes."""
        self.connection.send(["😀".encode(), "😀".encode()], text=True)
        # websockets sends an trailing empty fragment. That's an implementation detail.
        self.assertEqual(
            list(self.remote_connection.recv_streaming()),
            ["😀", "😀", ""],
        )

    def test_send_connection_closed_ok(self):
        """send raises ConnectionClosedOK after a normal closure."""
        self.remote_connection.close()
        with self.assertRaises(ConnectionClosedOK):
            self.connection.send("😀")

    def test_send_connection_closed_error(self):
        """send raises ConnectionClosedError after an error."""
        self.remote_connection.close(code=CloseCode.INTERNAL_ERROR)
        with self.assertRaises(ConnectionClosedError):
            self.connection.send("😀")

    def test_send_during_send(self):
        """send raises ConcurrencyError when called concurrently."""
        with self.run_in_thread(self.remote_connection.recv):
            send_gate = threading.Event()
            exit_gate = threading.Event()

            def fragments():
                yield "😀"
                send_gate.set()
                exit_gate.wait()
                yield "😀"

            with self.run_in_thread(self.connection.send, args=(fragments(),)):
                send_gate.wait()
                # The check happens in four code paths, depending on the argument.
                for message in [
                    "😀",
                    b"\x01\x02\xfe\xff",
                    ["😀", "😀"],
                    [b"\x01\x02", b"\xfe\xff"],
                ]:
                    with self.subTest(message=message):
                        with self.assertRaises(ConcurrencyError) as raised:
                            self.connection.send(message)
                        self.assertEqual(
                            str(raised.exception),
                            "cannot call send while another thread "
                            "is already running send",
                        )

                exit_gate.set()

    def test_send_empty_iterable(self):
        """send does nothing when called with an empty iterable."""
        self.connection.send([])
        self.connection.close()
        self.assertEqual(list(self.remote_connection), [])

    def test_send_mixed_iterable(self):
        """send raises TypeError when called with an iterable of inconsistent types."""
        with self.assertRaises(TypeError):
            self.connection.send(["😀", b"\xfe\xff"])

    def test_send_unsupported_iterable(self):
        """send raises TypeError when called with an iterable of unsupported type."""
        with self.assertRaises(TypeError):
            self.connection.send([None])

    def test_send_dict(self):
        """send raises TypeError when called with a dict."""
        with self.assertRaises(TypeError):
            self.connection.send({"type": "object"})

    def test_send_unsupported_type(self):
        """send raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            self.connection.send(None)

    # Test close.

    def test_close(self):
        """close sends a close frame."""
        self.connection.close()
        self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

    def test_close_explicit_code_reason(self):
        """close sends a close frame with a given code and reason."""
        self.connection.close(CloseCode.GOING_AWAY, "bye!")
        self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe9bye!"))

    def test_close_waits_for_close_frame(self):
        """close waits for a close frame then EOF before returning."""
        t0 = time.time()
        with self.delay_frames_rcvd(MS):
            self.connection.close()
        t1 = time.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    def test_close_waits_for_connection_closed(self):
        """close waits for EOF before returning."""
        if self.LOCAL is SERVER:
            self.skipTest("only relevant on the client-side")

        t0 = time.time()
        with self.delay_eof_rcvd(MS):
            self.connection.close()
        t1 = time.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    def test_close_timeout_waiting_for_close_frame(self):
        """close times out if no close frame is received."""
        t0 = time.time()
        with self.drop_frames_rcvd(), self.drop_eof_rcvd():
            self.connection.close()
        t1 = time.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.ABNORMAL_CLOSURE)
        self.assertGreater(t1 - t0, 2 * MS)

        with self.assertRaises(ConnectionClosedError) as raised:
            self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); no close frame received")
        self.assertIsInstance(exc.__cause__, TimeoutError)

    def test_close_timeout_waiting_for_connection_closed(self):
        """close times out if EOF isn't received."""
        if self.LOCAL is SERVER:
            self.skipTest("only relevant on the client-side")

        t0 = time.time()
        with self.drop_eof_rcvd():
            self.connection.close()
        t1 = time.time()

        self.assertEqual(self.connection.state, State.CLOSED)
        self.assertEqual(self.connection.close_code, CloseCode.NORMAL_CLOSURE)
        self.assertGreater(t1 - t0, 2 * MS)

        with self.assertRaises(ConnectionClosedOK) as raised:
            self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsInstance(exc.__cause__, TimeoutError)

    def test_close_preserves_queued_messages(self):
        """close preserves messages buffered in the assembler."""
        self.remote_connection.send("😀")
        self.connection.close()

        self.assertEqual(self.connection.recv(), "😀")
        with self.assertRaises(ConnectionClosedOK):
            self.connection.recv()

    def test_close_idempotency(self):
        """close does nothing if the connection is already closed."""
        self.connection.close()
        self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

        self.connection.close()
        self.assertNoFrameSent()

    def test_close_idempotency_race_condition(self):
        """close waits if the connection is already closing."""
        self.connection.close_timeout = 6 * MS

        def closer():
            with self.delay_frames_rcvd(4 * MS):
                self.connection.close()

        with self.run_in_thread(closer):
            #  run_in_thread() waits for MS, which lets closer() send a close frame.
            self.assertFrameSent(Frame(Opcode.CLOSE, b"\x03\xe8"))

            # Connection isn't closed yet.
            with self.assertRaises(TimeoutError):
                self.connection.recv(timeout=MS)

            self.connection.close()
            self.assertNoFrameSent()

            # Connection is closed now.
            with self.assertRaises(ConnectionClosedOK):
                self.connection.recv(timeout=MS)

    def test_close_during_recv(self):
        """close aborts recv when called concurrently with recv."""

        def closer():
            # Wait 2 * MS because run_in_thread() waits for MS.
            time.sleep(2 * MS)
            self.connection.close()

        with self.run_in_thread(closer):
            with self.assertRaises(ConnectionClosedOK) as raised:
                self.connection.recv()

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    def test_close_during_recv_streaming(self):
        """close aborts recv_streaming when called concurrently with recv_streaming."""

        def closer():
            # Wait 2 * MS because run_in_thread() waits for MS.
            time.sleep(2 * MS)
            self.connection.close()

        with self.run_in_thread(closer):
            with self.assertRaises(ConnectionClosedOK) as raised:
                for _ in self.connection.recv_streaming():
                    self.fail("did not raise")

        exc = raised.exception
        self.assertEqual(str(exc), "sent 1000 (OK); then received 1000 (OK)")
        self.assertIsNone(exc.__cause__)

    def test_close_during_send(self):
        """close fails the connection when called concurrently with send."""
        close_gate = threading.Event()
        exit_gate = threading.Event()

        def closer():
            close_gate.wait()
            self.connection.close()
            exit_gate.set()

        def fragments():
            yield "⏳"
            close_gate.set()
            exit_gate.wait()
            yield "⌛️"

        with self.run_in_thread(closer):
            iterator = fragments()
            with contextlib.closing(iterator):
                with self.assertRaises(ConnectionClosedError) as raised:
                    self.connection.send(iterator)

        exc = raised.exception
        self.assertEqual(
            str(exc),
            "sent 1011 (internal error) close during fragmented message; "
            "no close frame received",
        )
        self.assertIsNone(exc.__cause__)

    # Test ping.

    @patch("random.getrandbits")
    def test_ping(self, getrandbits):
        """ping sends a ping frame with a random payload."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping()
        getrandbits.assert_called_once_with(32)
        self.assertFrameSent(Frame(Opcode.PING, b"rand"))

    def test_ping_explicit_text(self):
        """ping sends a ping frame with a payload provided as text."""
        self.connection.ping("ping")
        self.assertFrameSent(Frame(Opcode.PING, b"ping"))

    def test_ping_explicit_binary(self):
        """ping sends a ping frame with a payload provided as binary."""
        self.connection.ping(b"ping")
        self.assertFrameSent(Frame(Opcode.PING, b"ping"))

    def test_acknowledge_ping(self):
        """ping is acknowledged by a pong with the same payload."""
        with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = self.connection.ping("this")
        self.remote_connection.pong("this")
        self.assertTrue(pong_received.wait(MS))

    def test_acknowledge_ping_non_matching_pong(self):
        """ping isn't acknowledged by a pong with a different payload."""
        with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = self.connection.ping("this")
        self.remote_connection.pong("that")
        self.assertFalse(pong_received.wait(MS))

    def test_acknowledge_previous_ping(self):
        """ping is acknowledged by a pong for as a later ping."""
        with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = self.connection.ping("this")
            self.connection.ping("that")
        self.remote_connection.pong("that")
        self.assertTrue(pong_received.wait(MS))

    def test_acknowledge_ping_on_close(self):
        """ping with ack_on_close is acknowledged when the connection is closed."""
        with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received_aoc = self.connection.ping("this", ack_on_close=True)
            pong_received = self.connection.ping("that")
        self.connection.close()
        self.assertTrue(pong_received_aoc.wait(MS))
        self.assertFalse(pong_received.wait(MS))

    def test_ping_duplicate_payload(self):
        """ping rejects the same payload until receiving the pong."""
        with self.drop_frames_rcvd():  # drop automatic response to ping
            pong_received = self.connection.ping("idem")

        with self.assertRaises(ConcurrencyError) as raised:
            self.connection.ping("idem")
        self.assertEqual(
            str(raised.exception),
            "already waiting for a pong with the same data",
        )

        self.remote_connection.pong("idem")
        self.assertTrue(pong_received.wait(MS))

        self.connection.ping("idem")  # doesn't raise an exception

    def test_ping_unsupported_type(self):
        """ping raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            self.connection.ping([])

    # Test pong.

    def test_pong(self):
        """pong sends a pong frame."""
        self.connection.pong()
        self.assertFrameSent(Frame(Opcode.PONG, b""))

    def test_pong_explicit_text(self):
        """pong sends a pong frame with a payload provided as text."""
        self.connection.pong("pong")
        self.assertFrameSent(Frame(Opcode.PONG, b"pong"))

    def test_pong_explicit_binary(self):
        """pong sends a pong frame with a payload provided as binary."""
        self.connection.pong(b"pong")
        self.assertFrameSent(Frame(Opcode.PONG, b"pong"))

    def test_pong_unsupported_type(self):
        """pong raises TypeError when called with an unsupported type."""
        with self.assertRaises(TypeError):
            self.connection.pong([])

    # Test keepalive.

    @patch("random.getrandbits")
    def test_keepalive(self, getrandbits):
        """keepalive sends pings at ping_interval and measures latency."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping_interval = 3 * MS
        self.connection.start_keepalive()
        self.assertIsNotNone(self.connection.keepalive_thread)
        self.assertEqual(self.connection.latency, 0)
        # 3 ms: keepalive() sends a ping frame.
        # 3.x ms: a pong frame is received.
        time.sleep(4 * MS)
        # 4 ms: check that the ping frame was sent.
        self.assertFrameSent(Frame(Opcode.PING, b"rand"))
        self.assertGreater(self.connection.latency, 0)
        self.assertLess(self.connection.latency, MS)

    def test_disable_keepalive(self):
        """keepalive is disabled when ping_interval is None."""
        self.connection.ping_interval = None
        self.connection.start_keepalive()
        self.assertIsNone(self.connection.keepalive_thread)

    @patch("random.getrandbits")
    def test_keepalive_times_out(self, getrandbits):
        """keepalive closes the connection if ping_timeout elapses."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping_interval = 4 * MS
        self.connection.ping_timeout = 2 * MS
        with self.drop_frames_rcvd():
            self.connection.start_keepalive()
            # 4 ms: keepalive() sends a ping frame.
            # 4.x ms: a pong frame is dropped.
            time.sleep(4 * MS)
            # Exiting the context manager sleeps for 1 ms.
        # 6 ms: no pong frame is received; the connection is closed.
        time.sleep(3 * MS)
        # 8 ms: check that the connection is closed.
        self.assertEqual(self.connection.state, State.CLOSED)

    @patch("random.getrandbits")
    def test_keepalive_ignores_timeout(self, getrandbits):
        """keepalive ignores timeouts if ping_timeout isn't set."""
        getrandbits.side_effect = itertools.count(1918987876)
        self.connection.ping_interval = 4 * MS
        self.connection.ping_timeout = None
        with self.drop_frames_rcvd():
            self.connection.start_keepalive()
            # 4 ms: keepalive() sends a ping frame.
            # 4.x ms: a pong frame is dropped.
            time.sleep(4 * MS)
            # Exiting the context manager sleeps for 1 ms.
        # 6 ms: no pong frame is received; the connection remains open.
        time.sleep(3 * MS)
        # 8 ms: check that the connection is still open.
        self.assertEqual(self.connection.state, State.OPEN)

    def test_keepalive_terminates_while_sleeping(self):
        """keepalive task terminates while waiting to send a ping."""
        self.connection.ping_interval = 3 * MS
        self.connection.start_keepalive()
        self.connection.keepalive_thread.join(MS)
        self.assertTrue(self.connection.keepalive_thread.is_alive())
        self.connection.close()
        self.connection.keepalive_thread.join(MS)
        self.assertFalse(self.connection.keepalive_thread.is_alive())

    def test_keepalive_terminates_when_sending_ping_fails(self):
        """keepalive task terminates when sending a ping fails."""
        self.connection.ping_interval = MS
        self.connection.start_keepalive()
        self.assertTrue(self.connection.keepalive_thread.is_alive())
        with self.drop_eof_rcvd(), self.drop_frames_rcvd():
            self.connection.close()
            # Exiting the context managers sleeps for 2 ms.
        self.assertFalse(self.connection.keepalive_thread.is_alive())

    def test_keepalive_terminates_while_waiting_for_pong(self):
        """keepalive task terminates while waiting to receive a pong."""
        self.connection.ping_interval = MS
        self.connection.ping_timeout = 4 * MS
        with self.drop_frames_rcvd():
            self.connection.start_keepalive()
            # 1 ms: keepalive() sends a ping frame.
            # 1.x ms: a pong frame is dropped.
            time.sleep(MS)
            # Exiting the context manager sleeps for 1 ms.
        # 2 ms: close the connection before ping_timeout elapses.
        self.connection.close()
        self.connection.keepalive_thread.join(MS)
        self.assertFalse(self.connection.keepalive_thread.is_alive())

    def test_keepalive_reports_errors(self):
        """keepalive reports unexpected errors in logs."""
        self.connection.ping_interval = 2 * MS
        self.connection.start_keepalive()
        # Inject a fault when waiting to receive a pong.
        with self.assertLogs("websockets", logging.ERROR) as logs:
            with patch("threading.Event.wait", side_effect=Exception("BOOM")):
                # 2 ms: keepalive() sends a ping frame.
                # 2.x ms: a pong frame is dropped.
                time.sleep(3 * MS)
        self.assertEqual(
            [record.getMessage() for record in logs.records],
            ["keepalive ping failed"],
        )
        self.assertEqual(
            [str(record.exc_info[1]) for record in logs.records],
            ["BOOM"],
        )

    # Test parameters.

    def test_close_timeout(self):
        """close_timeout parameter configures close timeout."""
        connection = Connection(
            Mock(spec=socket.socket),
            Protocol(self.LOCAL),
            close_timeout=42 * MS,
        )
        self.assertEqual(connection.close_timeout, 42 * MS)

    def test_max_queue(self):
        """max_queue configures high-water mark of frames buffer."""
        connection = Connection(
            Mock(spec=socket.socket),
            Protocol(self.LOCAL),
            max_queue=4,
        )
        self.assertEqual(connection.recv_messages.high, 4)

    def test_max_queue_none(self):
        """max_queue disables high-water mark of frames buffer."""
        connection = Connection(
            Mock(spec=socket.socket),
            Protocol(self.LOCAL),
            max_queue=None,
        )
        self.assertEqual(connection.recv_messages.high, None)
        self.assertEqual(connection.recv_messages.high, None)

    def test_max_queue_tuple(self):
        """max_queue configures high-water and low-water marks of frames buffer."""
        connection = Connection(
            Mock(spec=socket.socket),
            Protocol(self.LOCAL),
            max_queue=(4, 2),
        )
        self.assertEqual(connection.recv_messages.high, 4)
        self.assertEqual(connection.recv_messages.low, 2)

    # Test attributes.

    def test_id(self):
        """Connection has an id attribute."""
        self.assertIsInstance(self.connection.id, uuid.UUID)

    def test_logger(self):
        """Connection has a logger attribute."""
        self.assertIsInstance(self.connection.logger, logging.LoggerAdapter)

    @patch("socket.socket.getsockname")
    def test_local_address(self, getsockname):
        """Connection has a local_address attribute."""
        getsockname.return_value = ("sock", 1234)
        self.assertEqual(self.connection.local_address, ("sock", 1234))
        getsockname.assert_called_with()

    @patch("socket.socket.getpeername")
    def test_remote_address(self, getpeername):
        """Connection has a remote_address attribute."""
        getpeername.return_value = ("peer", 1234)
        self.assertEqual(self.connection.remote_address, ("peer", 1234))
        getpeername.assert_called_with()

    def test_state(self):
        """Connection has a state attribute."""
        self.assertIs(self.connection.state, State.OPEN)

    def test_request(self):
        """Connection has a request attribute."""
        self.assertIsNone(self.connection.request)

    def test_response(self):
        """Connection has a response attribute."""
        self.assertIsNone(self.connection.response)

    def test_subprotocol(self):
        """Connection has a subprotocol attribute."""
        self.assertIsNone(self.connection.subprotocol)

    def test_close_code(self):
        """Connection has a close_code attribute."""
        self.assertIsNone(self.connection.close_code)

    def test_close_reason(self):
        """Connection has a close_reason attribute."""
        self.assertIsNone(self.connection.close_reason)

    # Test reporting of network errors.

    def test_writing_in_recv_events_fails(self):
        """Error when responding to incoming frames is correctly reported."""
        # Inject a fault by shutting down the socket for writing — but not by
        # closing it because that would terminate the connection.
        self.connection.socket.shutdown(socket.SHUT_WR)
        # Receive a ping. Responding with a pong will fail.
        self.remote_connection.ping()
        with self.assertRaises(ConnectionClosedError) as raised:
            self.connection.recv()
        self.assertIsInstance(raised.exception.__cause__, BrokenPipeError)

    def test_writing_in_send_context_fails(self):
        """Error when sending outgoing frame is correctly reported."""
        # Inject a fault by shutting down the socket for writing — but not by
        # closing it because that would terminate the connection.
        self.connection.socket.shutdown(socket.SHUT_WR)
        # Sending a pong will fail.
        with self.assertRaises(ConnectionClosedError) as raised:
            self.connection.pong()
        self.assertIsInstance(raised.exception.__cause__, BrokenPipeError)

    # Test safety nets — catching all exceptions in case of bugs.

    @patch("websockets.protocol.Protocol.events_received", side_effect=AssertionError)
    def test_unexpected_failure_in_recv_events(self, events_received):
        """Unexpected internal error in recv_events() is correctly reported."""
        self.remote_connection.send("😀")
        # Reading the message will trigger the injected fault.
        with self.assertRaises(ConnectionClosedError) as raised:
            self.connection.recv()
        self.assertIsInstance(raised.exception.__cause__, AssertionError)

    @patch("websockets.protocol.Protocol.send_text", side_effect=AssertionError)
    def test_unexpected_failure_in_send_context(self, send_text):
        """Unexpected internal error in send_context() is correctly reported."""
        # Sending a message will trigger the injected fault.
        with self.assertRaises(ConnectionClosedError) as raised:
            self.connection.send("😀")
        self.assertIsInstance(raised.exception.__cause__, AssertionError)


class ServerConnectionTests(ClientConnectionTests):
    LOCAL = SERVER
    REMOTE = CLIENT



#!/usr/bin/env python

import asyncio
import signal

from websockets.asyncio.client import connect

async def client():
    async with connect("ws://localhost:8765") as websocket:
        # Close the connection when receiving SIGTERM.
        loop = asyncio.get_running_loop()
        def close():
            return loop.create_task(websocket.close())
        loop.add_signal_handler(signal.SIGTERM, close)

        # Process messages received on the connection.
        async for message in websocket:
            ...

asyncio.run(client())



from __future__ import annotations

import base64
import hashlib
import secrets
import sys

from .typing import BytesLike


__all__ = ["accept_key", "apply_mask"]


GUID = "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"


def generate_key() -> str:
    """
    Generate a random key for the Sec-WebSocket-Key header.

    """
    key = secrets.token_bytes(16)
    return base64.b64encode(key).decode()


def accept_key(key: str) -> str:
    """
    Compute the value of the Sec-WebSocket-Accept header.

    Args:
        key: Value of the Sec-WebSocket-Key header.

    """
    sha1 = hashlib.sha1((key + GUID).encode()).digest()
    return base64.b64encode(sha1).decode()


def apply_mask(data: BytesLike, mask: bytes | bytearray) -> bytes:
    """
    Apply masking to the data of a WebSocket message.

    Args:
        data: Data to mask.
        mask: 4-bytes mask.

    """
    if len(mask) != 4:
        raise ValueError("mask must contain 4 bytes")

    data_int = int.from_bytes(data, sys.byteorder)
    mask_repeated = mask * (len(data) // 4) + mask[: len(data) % 4]
    mask_int = int.from_bytes(mask_repeated, sys.byteorder)
    return (data_int ^ mask_int).to_bytes(len(data), sys.byteorder)
